"""
This module provides convenience functions that can be used throughout
the codebase.
"""

import re
import logging
from html.parser import HTMLParser
from html import escape, unescape
from djehuty.utils.constants import allowed_html_tags

class HTMLStripper (HTMLParser):
    """Overriden HTMLParser to strip HTML tags inspired by Django's implementation."""

    def __init__ (self):
        super().__init__(convert_charrefs=False)
        self.reset()
        self.state = []

    def handle_data (self, data):
        self.state.append (data)

    def handle_entityref (self, name):
        self.state.append (f"&{name};")

    def handle_charref (self, name):
        self.state.append (f"&#{name};")

    def get_data (self):
        """Return stripped HTML."""
        return "".join(self.state)


def html_to_plaintext (value, respect_newlines=False):
    """
    Outputs plain text without HTML tags.
    When RESPECT_NEWLINES is set to True, it will convert line
    breaks to newline characters.
    """
    if respect_newlines:
        value = value.replace("</p>", "</p>\n\n")
        value = value.replace("<br>", "\n")
        value = value.replace("<br/>", "\n")
        value = value.replace("<br />", "\n")

    while "<" in value and ">" in value:
        tag_count = value.count("<")
        html = HTMLStripper()
        html.feed (value)
        html.close()
        value = html.get_data()
        if tag_count == value.count("<"):
            break

    return value

def contains_disallowed_html (value):
    """Return True when there is a disallowed tag in VALUE, False otherwise."""

    if value is None:
        return False

    stripped_value = unescape (value)
    for tag in allowed_html_tags:
        stripped_value = stripped_value.replace (f"<{tag}>", "")
        stripped_value = stripped_value.replace (f"</{tag}>", "")

    return html_to_plaintext (stripped_value) != stripped_value

def encode_html (value, allow_simple_tags=True, allow_some_punctuation=True):
    """
    Returns VALUE but with encoded HTML entities.  When ALLOW_SIMPLE_TAGS is
    True, it doesn't encode HTML tags in constants.allowed_html_tags.
    """
    encoded_value = escape (value)
    if allow_simple_tags:
        for tag in allowed_html_tags:
            encoded_value = encoded_value.replace (f"&lt;{tag}&gt;", f"<{tag}>")
            encoded_value = encoded_value.replace (f"&lt;/{tag}&gt;", f"</{tag}>")

    if allow_some_punctuation:
        for character in [("&", "&amp;"), ("\"", "&quot;"), (" ", "&nbsp;")]:
            encoded_value = encoded_value.replace (character[1], character[0])

    return encoded_value

def value_or (record, key, other):
    """Return the value of KEY or OTHER."""
    try:
        return record[key]
    except (IndexError, KeyError, TypeError):
        return other

def value_or_none (record, key):
    """Return the value of KEY or None."""
    return value_or (record, key, None)

def pretty_print_size (num_bytes):
    """Return pretty-printed file size."""

    output = ""
    if not isinstance(num_bytes, int):
        output = "0B"
    elif num_bytes < 1000:
        output = f"{num_bytes}B"
    elif num_bytes < 1000000:
        output = f"{num_bytes/1000:.2f}KB"
    elif num_bytes < 1000000000:
        output = f"{num_bytes/1000000:.2f}MB"
    elif num_bytes < 1000000000000:
        output = f"{num_bytes/1000000000:.2f}GB"
    elif num_bytes < 1000000000000000:
        output = f"{num_bytes/1000000000000:.2f}TB"
    else:
        output = f"{num_bytes/1000000000000000:.2f}PB"

    return output

def opendap_sizes_to_bytes (size, units):
    """Return the bytes for a pretty-printed SIZE with UNITS."""
    output = size

    if units == "Pbytes":
        output = size * 1000000000000000
    elif units == "Tbytes":
        output = size * 1000000000000
    elif units == "Gbytes":
        output = size * 1000000000
    elif units == "Mbytes":
        output = size * 1000000
    elif units == "Kbytes":
        output = size * 1000

    return output

def decimal_coord(raw_input, axis, digits=5):
    '''
    Converts txt to string with decimal coordinates or None if invalid.
    Accepts input like "42.597" or "5º 38’ 18.5’’ E".
    axis is 'N' or 'E'
    5 digits is accuracy (RMS) of 40 cm.
    '''
    pattern = r"""^(-?\d+)[º°]\s*(\d+)[’']\s*((\d+)(\.\d?)?)((’’)|('')|"|”)\s*([NESW]?)$"""
    if raw_input is not None:
        try:
            deg = float(raw_input)
        except ValueError:
            raw_input = str(raw_input).strip()
            match = re.search(pattern, raw_input)
            if match:
                components = match.groups()
                deg = int(components[0]) + int(components[1])/60 + float(components[2])/3600
                direction = components[-1] #direction may be N,E,S,W, axis N,E.
                if (direction == 'S' and axis == 'N') or (direction == 'W' and axis == 'E'):
                    deg = -deg
                elif not direction in (axis, ''):
                    return None #something is wrong, probably mixed up lat and lon
            else:
                return None #something is wrong, undecipherable gibberish
        arc_rel = deg/90 if axis == 'N' else deg/180
        if abs(arc_rel) <= 1.:
            return f'{deg:.{digits}f}'

    return None

def decimal_coords(lat, lon, digits=5):
    '''
    Converts raw input lat, lon to decimal coordinats or None if invalid.
    '''
    lat_validated = decimal_coord(lat, 'N', digits=digits)
    lon_validated = decimal_coord(lon, 'E', digits=digits)
    return (lat_validated, lon_validated)

def self_or_value_or_none(record, key):
    '''
    Return record[key]['value'] or record[key] or none.
    Use: deal with triples where table2rdf transformed a string into a number.
    '''
    if key in record:
        return value_or (record[key], 'value', record[key])

    return None

def parses_to_int (input_string):
    """Return True when wrapping in int() would succeed, False otherwise."""
    try:
        int(input_string)
    except (ValueError, TypeError):
        return False

    return True

def deduplicate_list (alist):
    """
    Return deduplicated list, retaining original ordering,
    based on first occurrence of duplicates.
    """
    try:
        return list({item[1]:item[0] for item in list(enumerate(alist))})
    except TypeError:
        logging.error('Wrong type of %s in deduplicate_list', alist)
        return None

def make_citation (authors, year, title, version, item_type, doi,
                   publisher='4TU.ResearchData', max_cited_authors=5):
    """Return citation in standard Datacite format."""
    try:
        auths = [{key:val for key,val in author.items() if val}
                 for author in authors]   #remove empty name parts
        citation = '; '.join([
            (f"{author['last_name']}, {author['first_name']}" if
             not {'first_name','last_name'}-set(author)
             else value_or (author, 'full_name', "Unknown")) for author in auths[:max_cited_authors]])
        if authors[max_cited_authors:max_cited_authors+1]:
            citation += ' et. al.'
        citation += f' ({year}): {title}'
        if not citation.endswith('.'):
            citation += '.'
        citation += f' Version {version}. {publisher}. {item_type}. https://doi.org/{doi}'
        return citation
    except TypeError:
        logging.error('could not make citation for %s', doi)
        return None

def custom_field_name (name):
    """Return a predictable name for a custom field."""

    name = name.lower().replace(" ", "_")

    ## Exceptions to the custom field names.
    if name == "licence_remarks":
        name = "license_remarks"
    if name == "geolocation_latitude":
        name = "latitude"
    if name == "geolocation_longitude":
        name = "longitude"

    return name

def is_opendap_url (url):
    """Returns True when URL links to 4TU's OPeNDAP server, False otherwise."""
    try:
        return url.split("/")[2] == "opendap.4tu.nl"
    except (AttributeError, IndexError, KeyError):
        return False

def add_logging_level (level_name, level_number, method_name=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `level_name` becomes an attribute of the `logging` module with the value
    `level_number`. `method_name` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `method_name` is not specified, `level_name.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    not overwrite existing attributes.

    Code copied and adapted from the following Stack Overflow post:
    https://stackoverflow.com/questions/2183233/35804945#35804945
    """
    if not method_name:
        method_name = level_name.lower()

    if (hasattr(logging, level_name) or
        hasattr(logging, method_name) or
        hasattr(logging.getLoggerClass(), method_name)):
        return

    # This method was inspired by the answers to Stack Overflow post
    # http://stackoverflow.com/q/2183233/2988730, especially
    # http://stackoverflow.com/a/13638084/2988730
    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_number):
            self._log(level_number, message, args, **kwargs) # pylint: disable=protected-access

    def log_to_root(message, *args, **kwargs):
        logging.log(level_number, message, *args, **kwargs)

    logging.addLevelName(level_number, level_name)
    setattr(logging, level_name, level_number)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)

def landing_page_url (item_id, version=None, item_type="dataset", base_url=""):
    """Returns (a version-specific) URL to an item's landing page."""
    url = f"{base_url}/{item_type}s/{item_id}"
    if version:
        url += f"/{version}"
    return url

def split_author_name(name):
    """ Procedure to split name of author into first and last name.
        Works for often occurring name patterns:
        * first_name (initials) last_name
        * first_name I.N.I.T. last_name
        * first_name last_name
        * last_name """
    name = re.sub(r'\s+', ' ', name)
    if ')' in name:
        parts = name.split(')', 1)
        parts[0] += ')'
    elif '.' in name:
        parts = name[::-1].split('.', 1)
        parts = [part[::-1] for part in parts][::-1]
        parts[0] += '.'
    else:
        parts = name.split(' ', 1)
    parts = [part.strip() for part in parts]
    parts = ([''] + parts)[-2:]
    return parts

def split_string (input_string, delimiter='\\s' , is_quoted=False, maxsplit=-1):
    """Splits a string by a delimiter character and strips whitespace."""
    if not isinstance(input_string, str) or input_string == "":
        return None
    if input_string.count(delimiter) == 0:
        return [input_string]
    regex_pattern = re.compile(fr'''((?:[^{delimiter}])+)''')
    if is_quoted:
        regex_pattern = re.compile(fr'''((?:[^{delimiter}"']|"[^"]*"|'[^']*')+)''')
    words = regex_pattern.split(input_string)[1::2]
    if maxsplit == 0:
        return [input_string]
    if 0 < maxsplit < len(words):
        tmp_words = []
        tmp_words = words[:maxsplit]
        tmp_words.append(delimiter.join(words[maxsplit:]))
        words = tmp_words
    words[:] = [word.strip() for word in words]
    words[:] = [word for word in words if word]
    if is_quoted:
        for index, word in enumerate(words):
            if word[0] == word[-1] and word[0] in ['"', "'"]:
                words[index] = word[1:-1]

    return words

def strip_string (input_string):
    """Removes whitespace from the beginning and end of a string."""
    if isinstance (input_string, str):
        return input_string.strip()
    return input_string

def normalize_identifier (value, prefix_url):
    """Procedure to rewrite URLs to URIs."""
    # Don't process invalid entries
    if not isinstance(value, str):
        return None
    # Don't store empty values.
    value = value.strip()
    if value == "":
        return None
    # Strip the URI prefix from identifiers.
    if value.startswith (prefix_url):
        return value[len(prefix_url):]

    return value

def normalize_orcid (orcid):
    """Procedure to make storing ORCID identifiers consistent."""
    return normalize_identifier (orcid, "https://orcid.org/")

def normalize_doi (doi):
    """Procedure to make storing DOIs consistent."""
    return normalize_identifier (normalize_identifier (doi, "https://doi.org/"), "doi.org/")



"""
This module provides a general cache mechanism to avoid duplicated queries
to the database server. Any object can be cached, as long as the object is
serializable by means of 'json.dumps' and deseralizeable by means of
'json.loads'.
"""

import glob
import os
import logging
import hashlib
import json

class CacheLayer:
    """This class provides the caching layer."""

    def __init__ (self, storage_path):
        self.storage     = storage_path
        self.log         = logging.getLogger(__name__)

    def make_key (self, input_string):
        """Procedure to turn 'input_string' into a short, unique identifier."""
        if input_string is None:
            return None

        md5 = hashlib.new ("md5", usedforsecurity=False)
        md5.update (input_string.encode('utf-8'))
        return md5.hexdigest()

    def cache_is_ready(self):
        """Procedure to set up and test the ability to cache."""
        if self.storage is None:
            return False

        try:
            os.makedirs(self.storage, mode=0o700, exist_ok=True)
            return os.path.isdir(self.storage)
        except PermissionError:
            pass

        return False

    def cached_value (self, prefix, key, is_raw=False):
        """Returns the cached value or None."""
        try:
            filename = os.path.join (self.storage, f"{prefix}_{key}")
            with open(filename, "r",
                      encoding = "utf-8") as cache_file:
                cached = cache_file.read()
                if is_raw:
                    return cached
                return json.loads (cached)
        except OSError:
            self.log.debug ("No cached response for %s.", key)
        except json.decoder.JSONDecodeError:
            self.log.error ("Possible cache corruption at %s.", filename)

        return None

    def cache_value (self, prefix, key, value, query=None, is_raw=False):
        """Procedure to store 'value' as a cache."""
        try:
            cache_filename = os.path.join (self.storage, f"{prefix}_{key}")
            cache_fd = os.open (cache_filename, os.O_WRONLY | os.O_CREAT, 0o600)
            with open(cache_fd, "w", encoding = "utf-8") as cache_file:
                if is_raw:
                    cache_file.write (value)
                else:
                    cache_file.write (json.dumps(value))
                if os.name != 'nt':
                    os.fchmod (cache_fd, 0o400)  # pylint: disable=no-member

            if query is not None:
                query_filename = os.path.join (self.storage, f"{prefix}_{key}.sparql")
                query_fd = os.open (query_filename, os.O_WRONLY | os.O_CREAT, 0o600)
                with open(query_fd, "w", encoding = "utf-8") as query_file:
                    query_file.write(query)
                    if os.name != 'nt':
                        os.fchmod (query_fd, 0o400)  # pylint: disable=no-member
        except OSError:
            self.log.warning ("Failed to save cache for %s.", key)

        return value

    def invalidate_by_prefix (self, prefix):
        """Procedure to remove all cache items belonging to 'prefix'."""
        for file_path in glob.glob(os.path.join(self.storage, f"{prefix}_*")):
            try:
                os.remove(file_path)
            except FileNotFoundError:
                self.log.warning ("Trying to remove %s multiple times.", file_path)

        return True

    def invalidate_all (self):
        """Procedure to remove all cache items."""

        if not isinstance(self.storage, str):
            return False

        if self.storage in ("", "/"):
            return False

        files = glob.glob(os.path.join(self.storage, "*"))
        self.log.info ("Removing %d files.", len(files))
        for file_path in files:
            try:
                os.remove(file_path)
            except FileNotFoundError:
                self.log.warning ("Trying to remove %s multiple times.", file_path)
            except IsADirectoryError:
                pass

        return True



"""This module provides a program-wide configurable state."""

class RuntimeConfiguration:  # pylint: disable=too-few-public-methods
    """This class implements the configurable properties for djehuty."""

    _instance = None
    def __new__ (cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__ (self):

        self.address                     = None
        self.port                        = None
        self.alternative_port            = None
        self.base_url                    = None
        self.log_file                    = None
        self.use_reloader                = None
        self.use_debugger                = None
        self.maximum_workers             = None
        self.transactions_directory      = "."
        self.clear_cache_on_start        = False
        self.s3_buckets                  = {}
        self.s3_cache_storage            = None
        self.storage_locations           = []
        self.storage                     = None
        self.secondary_storage           = None
        self.secondary_storage_quirks    = False
        self.endpoint                    = "http://127.0.0.1:8890/sparql"
        self.update_endpoint             = None
        self.state_graph                 = "https://data.4tu.nl/portal/self-test"
        self.privileges                  = {}
        self.thumbnail_storage           = None
        self.profile_images_storage      = None
        self.iiif_cache_storage          = None
        self.enable_query_audit_log      = False
        self.account_quotas              = {}
        self.group_quotas                = {}
        self.default_quota               = 5000000000
        self.depositing_domains          = []
        self.delay_inserting_log_entries = False
        self.export_directory            = "."
        self.site_name                   = ""
        self.site_description            = ""
        self.site_shorttag               = "4tu"
        self.support_email_address       = ""
        self.startup_timestamp           = 0
        self.allow_crawlers              = False
        self.in_production               = False
        self.in_preproduction            = False
        self.using_uwsgi                 = False
        self.maintenance_mode            = False
        self.sandbox_message_css         = ""
        self.sandbox_message             = False
        self.notice_message              = False
        self.show_portal_summary         = True
        self.show_institutions           = True
        self.show_science_categories     = True
        self.show_latest_datasets        = True
        self.disable_2fa                 = False
        self.enable_iiif                 = False
        self.disable_collaboration       = False
        self.automatic_login_email       = None
        self.handle_certificate_path     = None
        self.handle_certificate          = None
        self.handle_private_key_path     = None
        self.handle_private_key          = None
        self.handle_url                  = None
        self.handle_prefix               = None
        self.handle_index                = None
        self.small_footer                = (
            '<div id="footer-wrapper2"><p>This repository is powered by '
            '<a href="https://github.com/4TUResearchData/djehuty">djehuty</a> '
            'built for <a href="https://data.4tu.nl">4TU.ResearchData</a>.'
            '</p></div>'
        )
        self.large_footer                = self.small_footer
        self.orcid_client_id             = None
        self.orcid_client_secret         = None
        self.orcid_endpoint              = None
        self.orcid_read_public_token     = None
        self.identity_provider           = None
        self.ror_url                     = None
        self.saml_config_path            = None
        self.saml_config                 = None
        self.saml_attribute_email        = "urn:mace:dir:attribute-def:mail"
        self.saml_attribute_first_name   = "urn:mace:dir:attribute-def:givenName"
        self.saml_attribute_last_name    = "urn:mace:dir:attribute-def:sn"
        self.saml_attribute_common_name  = "urn:mace:dir:attribute-def:cn"
        self.saml_attribute_groups       = None
        self.saml_attribute_group_prefix = None
        self.sram_organization_api_token = None
        self.sram_collaboration_id       = None
        self.datacite_url                = None
        self.datacite_id                 = None
        self.datacite_password           = None
        self.datacite_prefix             = None
        self.ssi_psk                     = None
        self.menu                        = []
        self.colors                      = {
            "primary-color":            "#f49120",
            "primary-color-hover":      "#d26000",
            "primary-color-active":     "#9d4800",
            "primary-foreground-color": "#000000",
            "privilege-button-color":   "#fce3bf",
            "footer-background-color":  "#707070",
            "background-color":         "#ffffff"
        }

config = RuntimeConfiguration()



"""
This module provides constants that can be used throughout
the codebase.
"""
group_to_member = { 28586: 'delft',
                    28628: 'delft',
                    28589: 'eindhoven',
                    28631: 'eindhoven',
                    28592: 'twente',
                    28634: 'twente',
                    28595: 'wageningen' }
member_url_names = { 'delft'     : 'Delft_University_of_Technology',
                     'eindhoven' : 'Eindhoven_University_of_Technology',
                     'twente'    : 'University_of_Twente',
                     'wageningen': 'Wageningen_University_and_Research',
                     'other'     : 'Other_institutions' }

filetypes_by_extension = {
    "bash":       "Shell",
    "bib":        "TeX",
    "c":          "C",
    "cc":         "C++",
    "cpp":        "C++",
    "cs":         "C#",
    "css":        "CSS",
    "cxx":        "C++",
    "d":          "D",
    "f":          "Fortran",
    "f77":        "Fortran",
    "f90":        "Fortran",
    "for":        "Fortran",
    "geojson":    "JSON",
    "go":         "Go",
    "gohtml":     "Go",
    "h":          "C",
    "hpp":        "C++",
    "hs":         "Haskell",
    "html":       "HTML",
    "htm":        "HTML",
    "hxx":        "C++",
    "in":         "Automake",
    "ipynb":      "Jupyter Notebook",
    "java":       "Java",
    "jl":         "Julia",
    "js":         "JavaScript",
    "json":       "JSON",
    "m":          "Matlab",
    "mat":        "Matlab",
    "md":         "Markdown",
    "mts":        "TypeScript",
    "mysql":      "SQL",
    "php":        "PHP",
    "pl":         "Perl",
    "pm":         "Perl",
    "py":         "Python",
    "r":          "R",
    "rb":         "Ruby",
    "rs":         "Rust",
    "sh":         "Shell",
    "sparql":     "SPARQL",
    "sql":        "SQL",
    "sty":        "TeX",
    "tex":        "TeX",
    "ts":         "TypeScript",
    "tsx":        "TypeScript",
    "xml":        "XML",
    "xyz":        "Chemical XYZ",
    "yaml":       "YAML",
    "yml":        "YAML",
}

allowed_html_tags = ["p", "strong", "em", "u", "ol", "ul", "li", "code", "pre", "br", "sup", "sub", "h2", "h3"]



"""This module provides an interface to store data fetched by the 'figshare' module."""

import os
import ast
import logging
from datetime import datetime
from threading import Lock
from secrets import token_urlsafe
import json
import time
import re
from defusedxml import ElementTree
from rdflib import Graph, Literal, RDF, RDFS, XSD, URIRef
import requests
from requests.utils import requote_uri
from djehuty.utils.convenience import value_or, value_or_none
from djehuty.utils.convenience import custom_field_name, opendap_sizes_to_bytes
from djehuty.utils import rdf

class DatabaseInterface:
    """
    A class that serializes the output produced by the 'figshare'
    module as RDF.
    """

    def __init__(self):
        self.store            = Graph()
        self.lock_for_inserts = Lock()
        self.container_uris_lock = Lock()
        self.container_uris   = {}
        script_path = os.path.dirname(os.path.abspath(__file__))
        with open(f'{script_path}/resources/dois.json', 'r',
                  encoding = 'utf-8') as dois_file:
            self.extra_dois = json.load(dois_file)
        with open(f'{script_path}/resources/contributors_organizations.json',
                  'r', encoding = 'utf-8') as contributors_file:
            extra = json.load(contributors_file)
        self.extra_contributors_organizations = {
            item_type: { pid: dict(versions) for pid, versions in extra[item_type] }
            for item_type in extra
        }
        with open(f'{script_path}/resources/public_collections.json', 'r',
                  encoding = 'utf-8') as coll_file:
            self.public_collection_datasets = json.load(coll_file)

    def __get_from_url (self, url: str, headers, parameters):
        """Procedure to perform a GET request to a Figshare-compatible endpoint."""
        try:
            response = requests.get(url,
                                    headers = headers,
                                    params  = parameters,
                                    timeout = 60)
            if response.status_code == 200:
                return response.text

            logging.error("%s returned %d.", url, response.status_code)
            logging.error("Error message:\n---\n%s\n---", response.text)

        except requests.exceptions.ConnectionError:
            logging.error("Failed to connect to %s.", url)

        return False

    def __get_file_size_for_catalog (self, url, recurse=None):
        """Returns the file size for an OPeNDAP catalog."""

        if not url.startswith('https://opendap.4tu.nl/thredds'):
            logging.debug("Trying to get data size: %s is not an opendap catalog", url)
            return 0

        if recurse is None:
            no_recurse = ('/IDRA/', '/darelux/', '/zandmotor/meteohydro/xband/catalog', '/CF_Drinking_water/')
            recurse = True not in [no_recurse_fragment in url for no_recurse_fragment in no_recurse]

        total_filesize = 0
        metadata_url   = url.replace(".html", ".xml")
        metadata       = self.__get_from_url (metadata_url, {}, {})

        if not metadata:
            logging.error("Couldn't get file metadata for %s.", url)
            return total_filesize

        namespaces  = { "c": "http://www.unidata.ucar.edu/namespaces/thredds/InvCatalog/v1.0" }
        xml_root    = ElementTree.fromstring(metadata)
        references  = xml_root.findall(".//c:catalogRef", namespaces)

        ## Recursively handle directories.
        if not recurse:
            logging.debug("Catalog %s is not recursed, subcatalogs are linked to other datasets.", url)
        elif not references:
            logging.debug("Catalog %s does not contain subcatalogs.", url)
        else:
            for reference in references:
                suffix = reference.attrib["{http://www.w3.org/1999/xlink}href"]
                suburl = metadata_url.replace("catalog.xml", suffix)
                total_filesize += self.__get_file_size_for_catalog (suburl, recurse=True)

        ## Handle regular files.
        files          = xml_root.findall(".//c:dataSize", namespaces)
        if not files:
            if total_filesize == 0:
                logging.debug ("There are no files in %s.", url)
            return total_filesize

        for single_file in files:
            units = single_file.attrib["units"]
            size  = ast.literal_eval(single_file.text)
            total_filesize += opendap_sizes_to_bytes (size, units)

        return total_filesize

    def __load_resource_file (self, filename):
        script_path = os.path.dirname(os.path.abspath(__file__))
        data        = None
        try:
            with open(f"{script_path}/resources/{filename}", "r",
                      encoding = "utf-8") as data_file:
                data = json.load(data_file)
        except FileNotFoundError:
            logging.error("Could not load resource '%s'.", filename)

        return data

    def get_crossref_events_for_doi (self, dois, from_date, mail_to):
        """Returns a list of events for a list of DOIs."""
        #rate limit is 15 queries per second. I'm going for 10/s
        rate = .1
        #from_date 'yyyy-mm-dd'
        # dois is list of dois
        # events is list of dictionaries (one per event)
        events = []
        for doi in dois:
            query = f"https://api.eventdata.crossref.org/v1/events?mailto={mail_to}&obj-id={doi}&from-occurred-date={from_date}"
            time.sleep(rate)
            req = requests.get(query, timeout = 60)
            data = json.loads(req.content)
            for item in data["message"]["events"]:
                try:
                    event = {}
                    event['object']    = str(item["obj_id"])
                    event['source']    = str(item["source_id"])
                    event['timestamp'] = str(item["timestamp"])
                    event['type']      = str(item["relation_type_id"])
                    events.append(event)
                except KeyError:
                    logging.error("Failed to gather crossref data for %s.", doi)

        return events

    def record_uri (self, record_type, identifier_name, identifier):
        """
        Returns the URI for a record identified with IDENTIFIER_NAME and by
        IDENTIFIER or None if no such URI can be found.
        """
        if identifier is None:
            return None

        if isinstance(identifier, str):
            identifier = f"\"{identifier}\"^^<{str(XSD.string)}>"

        try:
            query    = (
                "SELECT ?uri WHERE { "
                f"?uri <{str(RDF.type)}> <{str(rdf.DJHT[record_type])}> ; "
                f"<{str(rdf.DJHT[identifier_name])}> {identifier} . }}"
            )

            results = self.store.query (query)
            return results.bindings[0]["uri"]

        except (KeyError, IndexError):
            return None

    def serialize (self):
        """Output the triplets in the graph  to stdout."""
        body = self.store.serialize(format="turtle")
        if isinstance(body, bytes):
            body = body.decode('utf-8')

        print(body)

    def insert_account (self, record):
        """Procedure to add an account record to GRAPH."""

        uri = rdf.unique_node ("account")
        with self.lock_for_inserts:
            self.store.add ((uri, RDF.type, rdf.DJHT["Account"]))

            institution_user_id = value_or (record, "institution_user_id", None)
            domain              = None
            if institution_user_id is not None:
                domain = institution_user_id.partition("@")[2]

            rdf.add (self.store, uri, rdf.DJHT["id"],                    value_or (record, "id", None),           XSD.integer)
            rdf.add (self.store, uri, rdf.DJHT["active"],                value_or (record, "active", False),      XSD.boolean)
            rdf.add (self.store, uri, rdf.DJHT["domain"],                domain,                                  XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["email"],                 value_or (record, "email", None),        XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["first_name"],            value_or (record, "first_name", None),   XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["last_name"],             value_or (record, "last_name", None),    XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["institution_user_id"],   institution_user_id,                     XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["institution_id"],        value_or (record, "institution_id", None))
            rdf.add (self.store, uri, rdf.DJHT["group_id"],              value_or (record, "group_id", None))
            rdf.add (self.store, uri, rdf.DJHT["pending_quota_request"], value_or (record, "pending_quota_request", None))
            rdf.add (self.store, uri, rdf.DJHT["used_quota_public"],     value_or (record, "used_quota_public", None))
            rdf.add (self.store, uri, rdf.DJHT["used_quota_private"],    value_or (record, "used_quota_private", None))
            rdf.add (self.store, uri, rdf.DJHT["used_quota"],            value_or (record, "used_quota", None))
            rdf.add (self.store, uri, rdf.DJHT["maximum_file_size"],     value_or (record, "maximum_file_size", None))
            rdf.add (self.store, uri, rdf.DJHT["quota"],                 value_or (record, "quota", None))
            rdf.add (self.store, uri, rdf.DJHT["modified_date"],         value_or_none (record, "modified_date"), XSD.dateTime)
            rdf.add (self.store, uri, rdf.DJHT["created_date"],          value_or_none (record, "created_date"),  XSD.dateTime)

        return uri

    def insert_institution (self, record):
        """Procedure to insert an institution record."""

        try:
            institution_id = record["institution_id"]
            uri            = rdf.ROW[f"institution_{institution_id}"]

            self.store.add ((uri, RDF.type,        rdf.DJHT["Institution"]))
            self.store.add ((uri, rdf.DJHT["id"],   Literal(institution_id, datatype=XSD.integer)))
            self.store.add ((uri, rdf.DJHT["name"], Literal(record["name"], datatype=XSD.string)))

            return True

        except KeyError:
            pass

        logging.error ("Failed to insert Institution record: %s", record)
        return False

    def insert_account_author_link (self, account_uri, author_uri):
        """Procedure to insert a link between an account and an author."""
        self.store.add ((author_uri, rdf.DJHT["account"], URIRef(account_uri)))
        return True

    def insert_author (self, record):
        """Procedure to insert an author record."""

        author_id = value_or_none (record, "id")
        if author_id is None:
            logging.error ("Invalid author record: %s", record)
            return None
        uri = self.record_uri ("Author", "id", author_id)
        if uri is not None:
            return uri

        uri       = rdf.unique_node ("author")
        is_active = "is_active" in record and record["is_active"]
        is_public = "is_public" in record and record["is_public"]

        self.store.add ((uri, RDF.type, rdf.DJHT["Author"]))

        rdf.add (self.store, uri, rdf.DJHT["id"],             author_id,                                 XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["institution_id"], value_or (record, "institution_id", None), XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["group_id"],       value_or (record, "group_id",       None), XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["first_name"],     value_or (record, "first_name",     None), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["last_name"],      value_or (record, "last_name",      None), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["full_name"],      value_or (record, "full_name",      None), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["job_title"],      value_or (record, "job_title",      None), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["url_name"],       value_or (record, "url_name",       None), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["orcid_id"],       value_or (record, "orcid_id",       None), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["is_active"],      is_active,                                 XSD.boolean)
        rdf.add (self.store, uri, rdf.DJHT["is_public"],      is_public,                                 XSD.boolean)

        return uri

    def insert_timeline (self, uri, record):
        """Procedure to insert a timeline record."""

        if not record:
            return False

        rdf.add (self.store, uri, rdf.DJHT["revision_date"],     value_or_none (record, "revision"),             XSD.dateTime)
        rdf.add (self.store, uri, rdf.DJHT["publisher_publication_date"], value_or_none (record, "publisherPublication"), XSD.dateTime)
        rdf.add (self.store, uri, rdf.DJHT["publisher_acceptance_date"], value_or_none (record, "publisherAcceptance"), XSD.dateTime)
        rdf.add (self.store, uri, rdf.DJHT["posted_date"],       value_or_none (record, "posted"),               XSD.dateTime)
        rdf.add (self.store, uri, rdf.DJHT["submission_date"],   value_or_none (record, "submission"),           XSD.dateTime)

        return True

    def insert_category (self, record):
        """Procedure to insert a category record."""

        category_id = value_or_none (record, "id")
        uri = self.record_uri ("Category", "id", category_id)
        if uri is not None:
            return uri

        uri = rdf.unique_node ("category")
        self.store.add ((uri, RDF.type,      rdf.DJHT["Category"]))
        self.store.add ((uri, rdf.DJHT["id"], Literal(category_id)))

        rdf.add (self.store, uri, rdf.DJHT["title"],       value_or (record, "title", None),       XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["parent_id"],   value_or (record, "parent_id", None),   XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["source_id"],   value_or (record, "source_id", None),   XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["taxonomy_id"], value_or (record, "taxonomy_id", None), XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["classification_code"], value_or (record, "classification_code", None), XSD.string)

        return uri

    def insert_record_list (self, uri, records, name, insert_procedure):
        """
        Adds an RDF list with indexes for RECORDS to the graph using
        INSERT_PROCEDURE.  The INSERT_PROCEDURE must take  a single item
        from RECORDS, and it must return the URI used as subject to describe
        the record.
        """
        if records:
            blank_node = rdf.blank_node ()
            self.store.add ((uri, rdf.DJHT[name], blank_node))

            previous_blank_node = None
            for index, item in enumerate(records):
                record_uri = insert_procedure (item)
                self.store.add ((blank_node, rdf.DJHT["index"], Literal (index, datatype=XSD.integer)))
                self.store.add ((blank_node, RDF.first,        record_uri))

                if previous_blank_node is not None:
                    self.store.add ((previous_blank_node, RDF.rest, blank_node))
                    self.store.add ((previous_blank_node, RDF.type, RDF.List))

                previous_blank_node = blank_node
                blank_node = rdf.blank_node ()

            del blank_node
            self.store.add ((previous_blank_node, RDF.rest, RDF.nil))
            self.store.add ((previous_blank_node, RDF.type, RDF.List))

    def insert_category_list (self, uri, categories):
        """Adds an RDF list with indexes for CATEGORIES to GRAPH."""
        self.insert_record_list (uri, categories, "categories", self.insert_category)

    def insert_author_list (self, uri, authors):
        """Adds an RDF list with indexes for AUTHORS to GRAPH."""
        self.insert_record_list (uri, authors, "authors", self.insert_author)

    def insert_file_list (self, uri, files):
        """Adds an RDF list with indexes for FILES to GRAPH."""
        self.insert_record_list (uri, files, "files", self.insert_file)

    def insert_funding_list (self, uri, records):
        """Adds an RDF list with indexes for FUNDINGS to GRAPH."""
        self.insert_record_list (uri, records, "funding_list", self.insert_funding)

    def insert_private_links_list (self, uri, records):
        """Adds an RDF list with indexes for PRIVATE_LINKS to GRAPH."""
        self.insert_record_list (uri, records, "private_links", self.insert_private_link)

    def insert_item_list (self, uri, items, items_name):
        """Adds an RDF list with indexes for ITEMS to GRAPH."""

        if items:
            blank_node = rdf.blank_node ()
            self.store.add ((uri, rdf.DJHT[items_name], blank_node))

            previous_blank_node = None
            for index, item in enumerate(items):
                self.store.add ((blank_node, rdf.DJHT["index"], Literal (index, datatype=XSD.integer)))
                if isinstance (item, URIRef):
                    self.store.add ((blank_node, RDF.first,        item))
                else:
                    self.store.add ((blank_node, RDF.first,        Literal (item,  datatype=XSD.string)))

                if previous_blank_node is not None:
                    self.store.add ((previous_blank_node, RDF.rest, blank_node))
                    self.store.add ((previous_blank_node, RDF.type, RDF.List))

                previous_blank_node = blank_node
                blank_node = rdf.blank_node ()

            del blank_node
            self.store.add ((previous_blank_node, RDF.rest, RDF.nil))
            self.store.add ((previous_blank_node, RDF.type, RDF.List))

    def insert_custom_field_value (self, uri, name, value, field_type):
        """Insert a single NAME-VALUE pair for URI into GRAPH."""

        # Custom fields can be either text or URLs.
        # URLs should be converted to URIRefs, while all other
        # cases should be considered text strings.
        if field_type != "url":
            field_type = XSD.string

        if isinstance (value, list):
            for item in value:
                if isinstance (item, str) and item == "":
                    continue

                # The URLs in the "Data Link" custom field aren't encoded.
                # This could and would lead to invalid RDF serialization.
                # So, we must re-encode the URLs to treat them as RDF URIs.
                if name == "data_link":
                    item = requote_uri (item)

                rdf.add (self.store, uri, rdf.DJHT[name], item, field_type)

        elif not (isinstance (value, str) and value == ""):
            if name == "data_link":
                value = requote_uri (value)
            rdf.add (self.store, uri, rdf.DJHT[name], value, field_type)

    def insert_custom_field (self, uri, field):
        """Procedure to insert a custom_field record."""

        name = custom_field_name (field["name"])

        settings    = {}
        validations = {}
        if "settings" in field:
            settings = field["settings"]
            if "validations" in settings:
                validations = settings["validations"]

        default_value = None
        if "default_value" in settings:
            default_value = settings["default_value"]

        ## Avoid inserting the custom field's properties multiple times.
        subjects = self.store.subjects ((None, RDF.type, rdf.DJHT["CustomField"]),
                                        (None, rdf.DJHT["name"], Literal(name, datatype=XSD.string)))
        field_uri = next (subjects, None)
        if field_uri is None:
            field_uri = rdf.ROW[f"custom_field_{name}"]
            self.store.add ((field_uri,     RDF.type,                rdf.DJHT["CustomField"]))
            rdf.add (self.store, field_uri, rdf.DJHT["predicate"],    rdf.DJHT[name], datatype="url")
            rdf.add (self.store, field_uri, rdf.DJHT["name"],         name,                                        XSD.string)
            rdf.add (self.store, field_uri, rdf.DJHT["original_name"], field["name"],                              XSD.string)
            rdf.add (self.store, field_uri, rdf.DJHT["max_length"],   value_or_none (validations, "max_length"))
            rdf.add (self.store, field_uri, rdf.DJHT["min_length"],   value_or_none (validations, "min_length"))
            rdf.add (self.store, field_uri, rdf.DJHT["is_mandatory"], value_or_none (validations, "is_mandatory"), XSD.boolean)
            rdf.add (self.store, field_uri, rdf.DJHT["placeholder"],  value_or_none (validations, "placeholder"),  XSD.string)
            rdf.add (self.store, field_uri, rdf.DJHT["is_multiple"],  value_or_none (validations, "is_multiple"),  XSD.boolean)

        if isinstance(field["value"], list):
            for value in field["value"]:
                value      = value_or (field, "value", default_value)
                field_type = value_or_none (field, "field_type")
                self.insert_custom_field_value (uri, name, value, field_type)

                ## Drop-down lists have predefined items to choose from.
                ## We must add those predefined items once to the RDF store.
                if field_type == "dropdown" and value_or (settings, "options", False):
                    options_uri = self.record_uri ("CustomFieldOption", "name", name)
                    if options_uri is None:
                        options_uri = rdf.ROW[f"custom_field_options_{name}"]
                        options     = value_or_none (settings, "options")
                        rdf.add (self.store, options_uri, rdf.DJHT["name"], name, XSD.string)
                        self.insert_item_list (options_uri, options, "values")
        else:
            value      = value_or (field, "value", default_value)
            field_type = value_or_none (field, "field_type")
            self.insert_custom_field_value (uri, name, value, field_type)

    def fix_doi (self, record, item_id, version, item_type):
        '''Fix doi if needed: handle Fedora/Figshare history and fix Figshare bug'''
        extra_dois = [extra['doi'] for extra in self.extra_dois[item_type] if extra['id']==item_id and extra['version']==version]
        if extra_dois:
            record['doi'] = extra_dois[0]
        elif version and 'doi' in record and re.findall(r'^10\.4121/(c\.)?\d+$', record['doi']):
            record['doi'] += f'.v{version}'

    def add_container_doi (self, record, container):
        '''Add doi to container if latest item in container has Figshare-style doi'''
        try:
            groups = re.match(r'^(10\.4121/(c\.)?\d+)(.v\d+)?$', record['doi']).groups()
            if groups:
                rdf.add (self.store, container, rdf.DJHT["doi"], groups[0], XSD.string)
        except (KeyError, IndexError, AttributeError):
            pass # No DOI in record or unnecessary to add to the container.

    def handle_custom_fields (self, record, uri, item_id, version, item_type):
        '''Handle custom fields and fix special cases'''
        for field in value_or (record, "custom_fields", []):
            #replace contributors by organizations and (optional) contributors
            if field['name'] == 'Contributors':
                try:
                    extra = self.extra_contributors_organizations[item_type][item_id][version]
                    contributors  = extra['contributors']
                    if contributors:
                        field['value'] = contributors
                        self.insert_custom_field (uri, field)
                    field = {'name': 'Organizations', 'value': extra['organizations']}
                except KeyError:
                    pass
            #add "Data Link Size" field for opendap catalogs in "Data Link" field, correct opendap url if needed
            if field['name'] == 'Data Link':
                urls = field['value']
                urls = [url.replace('https://opendap.tudelft.nl/', 'https://opendap.4tu.nl/') for url in urls]
                field['value'] = urls
                data_link_size = sum(self.__get_file_size_for_catalog(url) for url in urls)
                if data_link_size:
                    rdf.add (self.store, uri, rdf.DJHT["data_link_size"], data_link_size, XSD.integer)

            self.insert_custom_field (uri, field)

    def insert_collection (self, record, account_id, account_uri):
        """Procedure to insert a collection record."""

        collection_id  = record["id"]
        uri            = rdf.unique_node ("collection")

        is_public          = bool (value_or (record, "public", False))
        is_latest          = bool (value_or (record, "is_latest", False))
        is_editable        = bool (value_or (record, "is_editable", False))

        version = value_or_none(record, "version")
        self.fix_doi (record, collection_id, version, 'collection')

        with self.lock_for_inserts:
            self.store.add ((uri, RDF.type,                 rdf.DJHT["Collection"]))
            self.store.add ((uri, rdf.DJHT["collection_id"], Literal(collection_id, datatype=XSD.integer)))

            rdf.add (self.store, uri, rdf.DJHT["url"],                 value_or_none (record, "url"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["title"],               value_or_none (record, "title"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["published_date"],      value_or_none (record, "published_date"), XSD.dateTime)
            rdf.add (self.store, uri, rdf.DJHT["created_date"],        value_or_none (record, "created_date"), XSD.dateTime)
            rdf.add (self.store, uri, rdf.DJHT["modified_date"],       value_or_none (record, "modified_date"), XSD.dateTime)
            rdf.add (self.store, uri, rdf.DJHT["doi"],                 value_or_none (record, "doi"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["citation"],            value_or_none (record, "citation"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["group_id"],            value_or_none (record, "group_id"), XSD.integer)
            ## group_resource_id is always empty/NULL.
            #rdf.add (self.store, uri, rdf.DJHT["group_resource_id"],   value_or_none (record, "group_resource_id"), XSD.integer)
            rdf.add (self.store, uri, rdf.DJHT["institution_id"],      value_or_none (record, "institution_id"), XSD.integer)
            rdf.add (self.store, uri, rdf.DJHT["description"],         value_or_none (record, "description"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["version"],             value_or_none (record, "version"), XSD.integer)
            ## resource_id is always empty/NULL.
            #rdf.add (self.store, uri, rdf.DJHT["resource_id"],         value_or_none (record, "resource_id"), XSD.integer)
            rdf.add (self.store, uri, rdf.DJHT["resource_doi"],        value_or_none (record, "resource_doi"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["resource_title"],      value_or_none (record, "resource_title"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["resource_version"],    value_or_none (record, "resource_version"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["resource_link"],       value_or_none (record, "resource_link"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["handle"],              value_or_none (record, "handle"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["is_public"],           is_public, XSD.boolean)
            rdf.add (self.store, uri, rdf.DJHT["is_latest"],           is_latest, XSD.boolean)
            rdf.add (self.store, uri, rdf.DJHT["is_editable"],         is_editable, XSD.boolean)

            self.insert_timeline (uri, value_or_none (record, "timeline"))
            self.insert_author_list (uri, value_or (record, "authors", []))
            self.insert_category_list (uri, value_or (record, "categories", []))
            self.insert_funding_list (uri, value_or (record, "funding_list", []))
            self.insert_private_links_list (uri, value_or (record, "private_links", []))
            tags = [tag for tag in value_or (record, "tags", []) if not tag.startswith('Collection: ')]
            self.insert_item_list (uri, tags, "tags")
            self.insert_item_list (uri, value_or (record, "references", []), "references")

            self.handle_custom_fields (record, uri, collection_id, version, 'collections')

            datasets = value_or (record, "datasets",
                                 value_or (self.public_collection_datasets,
                                           f"{collection_id}_{version}", []))
            if datasets:
                for index, dataset_id in enumerate (datasets):
                    dataset_uri = self.record_uri ("DatasetContainer", "dataset_id", dataset_id)
                    if dataset_uri is None:
                        logging.error ("Could not find dataset container for %d", dataset_id)
                        continue

                    datasets[index] = URIRef (dataset_uri)

                self.insert_item_list (uri, datasets, "datasets")
            else:
                logging.warning ("Collection %d seems to be empty.", collection_id)

            ## Assign the collection to the container
            container = self.container_uri (collection_id, "collection",
                                            account_id, account_uri)
            rdf.add (self.store, uri, rdf.DJHT["container"], container, datatype="uri")

            if "statistics" in record:
                stats = record["statistics"]
                self.insert_totals_statistics (stats["totals"], container)
            elif is_public and is_latest:
                logging.warning ("No statistics available for collection %d.", collection_id)

            if is_editable:
                self.store.add ((container, rdf.DJHT["draft"], uri))
            else:
                new_blank_node = rdf.blank_node ()
                self.store.add ((new_blank_node, RDF.type, RDF.List))
                self.store.add ((new_blank_node, RDF.first, uri))
                self.store.add ((new_blank_node, RDF.rest, RDF.nil))

                blank_node = self.last_list_node (container, "published_versions")
                if blank_node is None:
                    self.store.add    ((container, rdf.DJHT["published_versions"], new_blank_node))
                else:
                    self.store.remove ((blank_node, RDF.rest, RDF.nil))
                    self.store.add    ((blank_node, RDF.rest, new_blank_node))

            if is_latest:
                self.store.add ((container, rdf.DJHT["latest_published_version"], uri))
                self.add_container_doi (record, container)
                timeline = value_or_none (record, "timeline")
                rdf.add (self.store, container, rdf.DJHT["first_online_date"],
                         value_or_none (timeline, "firstOnline"), XSD.dateTime)

        return True

    def insert_funding (self, record):
        """Procedure to insert a funding record."""

        funding_id = value_or_none (record, "id")
        uri = self.record_uri ("Funding", "id", funding_id)
        if uri is not None:
            return uri

        is_user_defined = value_or (record, "is_user_defined", False)

        uri = rdf.unique_node ("funding")
        self.store.add ((uri, RDF.type, rdf.DJHT["Funding"]))
        rdf.add (self.store, uri, rdf.DJHT["id"],              funding_id, XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["title"],           value_or_none (record, "title"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["grant_code"],      value_or_none (record, "grant_code"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["funder_name"],     value_or_none (record, "funder_name"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["url"],             value_or_none (record, "url"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["is_user_defined"], is_user_defined, XSD.boolean)

        return uri

    def insert_private_link (self, record):
        """Procedure to insert a private link to a dataset or a collection."""

        is_active = value_or (record, "is_active", False)
        suffix    = token_urlsafe (64)
        uri       = rdf.unique_node ("private_link")

        self.store.add ((uri, RDF.type, rdf.DJHT["PrivateLink"]))
        rdf.add (self.store, uri, rdf.DJHT["id"],           value_or_none (record, "id"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["suffix"],       suffix, XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["expires_date"], value_or_none (record, "expires_date"), XSD.dateTime)
        rdf.add (self.store, uri, rdf.DJHT["is_active"],    is_active, XSD.boolean)

        return uri

    def insert_license (self, uri, record):
        """Procedure to insert a license record."""

        if not (record and value_or (record, "url", False)):
            return False

        license_uri  = URIRef (record["url"])

        ## Insert the license if it isn't in the graph.
        if (license_uri, RDF.type, rdf.DJHT["License"]) not in self.store:
            license_name = value_or_none (record,"name")
            license_type = value_or_none (record,"type")
            license_id   = value_or_none (record,"value")
            license_spdx = value_or_none (record, "spdx")
            self.store.add ((license_uri, RDF.type, rdf.DJHT["License"]))
            self.store.add ((license_uri, rdf.DJHT["name"],
                             Literal(license_name, datatype=XSD.string)))
            self.store.add ((license_uri, rdf.DJHT["id"],
                             Literal(license_id, datatype=XSD.integer)))

            if license_spdx is not None:
                self.store.add ((license_uri, rdf.DJHT["spdx"],
                                 URIRef(license_spdx)))

            license_type_uri = None
            if license_type == "software":
                license_type_uri = rdf.DJHT["SoftwareLicense"]
            if license_type == "hardware":
                license_type_uri = rdf.DJHT["HardwareLicense"]
            elif license_type == "data":
                license_type_uri = rdf.DJHT["DataLicense"]
            elif license_type == "legacy":
                license_type_uri = rdf.DJHT["LegacyLicense"]

            rdf.add (self.store, license_uri, rdf.DJHT["type"], license_type_uri, "url")
            rdf.add (self.store, license_type_uri, RDFS.label,  license_type, XSD.string)

        ## Insert the link between URI and the license.
        if uri is not None:
            self.store.add ((uri, rdf.DJHT["license"], license_uri))

        return True

    def insert_totals_statistics (self, record, uri):
        """Procedure to insert simplified totals for a dataset or collection."""

        if record is None:
            return None

        now = datetime.strftime (datetime.now(), "%Y-%m-%dT%H:%M:%SZ")
        rdf.add (self.store, uri, rdf.DJHT["total_views"],     value_or_none (record, "views"), XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["total_downloads"], value_or_none (record, "downloads"), XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["total_shares"],    value_or_none (record, "shares"), XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["total_cites"],     value_or_none (record, "cites"), XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["totals_created_at"], now, XSD.dateTime)

        return True

    def insert_file (self, record):
        """Procedure to insert a file record."""

        file_id = value_or_none (record, "id")
        uri = self.record_uri ("File", "id", file_id)
        if uri is not None:
            return uri

        if record["download_url"].startswith("https://opendap.tudelft.nl/thredds"):
            record["download_url"] = record["download_url"].replace("opendap.tudelft.nl",
                                                                    "opendap.4tu.nl")

        if (record["download_url"].startswith("https://opendap.4tu.nl/thredds") and
            record["size"] == 0):
            record["size"] = self.__get_file_size_for_catalog (record["download_url"])

        is_link_only = bool (value_or (record, "is_link_only", False))

        uri = rdf.unique_node ("file")
        self.store.add ((uri, RDF.type, rdf.DJHT["File"]))

        rdf.add (self.store, uri, rdf.DJHT["id"],            file_id, XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["name"],          value_or_none (record, "name"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["size"],          value_or_none (record, "size"), XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["download_url"],  value_or_none (record, "download_url"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["supplied_md5"],  value_or_none (record, "supplied_md5"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["computed_md5"],  value_or_none (record, "computed_md5"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["viewer_type"],   value_or_none (record, "viewer_type"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["preview_state"], value_or_none (record, "preview_state"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["status"],        value_or_none (record, "status"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["upload_url"],    value_or_none (record, "upload_url"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["upload_token"],  value_or_none (record, "upload_token"), XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["is_link_only"],  is_link_only, XSD.boolean)

        return uri

    def last_list_node (self, parent_uri, predicate_name):
        """Returns the URI of the last blank node for PARENT_URI or None."""

        try:
            predicate = str(rdf.DJHT[predicate_name])
            query     = (
                "SELECT ?uri WHERE { "
                f"<{str(parent_uri)}> <{predicate}>/<{str(RDF.rest)}>* ?uri . "
                f"?uri <{str(RDF.rest)}> <{str(RDF.nil)}> . }}"
            )

            results  = self.store.query (query)
            return results.bindings[0]["uri"]

        except KeyError:
            pass
        except IndexError:
            pass

        return None

    def container_uri (self, item_id, item_type, account_id, account_uri):
        """Returns the URI of the dataset container belonging to dataset_id."""

        prefix     = item_type.capitalize()
        item_class = f"{prefix}Container"
        key        = f"{item_type}:{item_id}"
        uri = None
        with self.container_uris_lock:
            try:
                uri = self.container_uris[key]
            except KeyError:
                uri = self.record_uri (item_class, f"{item_type}_id", item_id)
                self.container_uris[key] = uri

        if uri is None:
            uri = rdf.unique_node ("container")
            self.store.add ((uri, RDF.type,                   rdf.DJHT[item_class]))
            self.store.add ((uri, rdf.DJHT[f"{item_type}_id"], Literal(item_id, datatype=XSD.integer)))
            self.store.add ((uri, rdf.DJHT["account_id"],      Literal(account_id, datatype=XSD.integer)))
            self.store.add ((uri, rdf.DJHT["account"],         URIRef(account_uri)))

            with self.container_uris_lock:
                self.container_uris[key] = uri

        return uri

    def insert_dataset (self, record):
        """Procedure to insert a dataset record."""

        dataset_id = value_or_none (record, "id")
        if dataset_id is None:
            return False
        version = value_or_none(record, "version")
        self.fix_doi (record, dataset_id, version, 'article')

        account_id = value_or_none (record, "account_id")
        account_uri = value_or_none (record, "account_uri")
        uri        = rdf.unique_node ("dataset")

        is_embargoed       = bool (value_or (record, "is_embargoed", False))
        is_public          = bool (value_or (record, "is_public", False))
        is_latest          = bool (value_or (record, "is_latest", False))
        is_editable        = bool (value_or (record, "is_editable", False))
        is_confidential    = bool (value_or (record, "is_confidential", False))
        is_metadata_record = bool (value_or (record, "is_metadata_record", False))
        has_linked_file    = bool (value_or (record, "has_linked_file", False))

        with self.lock_for_inserts:
            self.store.add ((uri, RDF.type,              rdf.DJHT["Dataset"]))
            self.store.add ((uri, rdf.DJHT["dataset_id"], Literal(dataset_id, datatype=XSD.integer)))

            self.insert_timeline (uri, value_or_none (record, "timeline"))
            self.insert_license (uri, value_or_none (record, "license"))

            rdf.add (self.store, uri, rdf.DJHT["title"],               value_or_none (record, "title"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["doi"],                 value_or_none (record, "doi"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["handle"],              value_or_none (record, "handle"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["group_id"],            value_or_none (record, "group_id"), XSD.integer)
            rdf.add (self.store, uri, rdf.DJHT["url"],                 value_or_none (record, "url"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["url_public_html"],     value_or_none (record, "url_public_html"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["url_public_api"],      value_or_none (record, "url_public_api"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["url_private_html"],    value_or_none (record, "url_private_html"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["url_private_api"],     value_or_none (record, "url_private_api"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["published_date"],      value_or_none (record, "published_date"), XSD.dateTime)
            rdf.add (self.store, uri, rdf.DJHT["created_date"],        value_or_none (record, "created_date"), XSD.dateTime)
            rdf.add (self.store, uri, rdf.DJHT["modified_date"],       value_or_none (record, "modified_date"), XSD.dateTime)
            rdf.add (self.store, uri, rdf.DJHT["thumb"],               value_or_none (record, "thumb"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["defined_type"],        value_or_none (record, "defined_type"), XSD.integer)
            rdf.add (self.store, uri, rdf.DJHT["defined_type_name"],   value_or_none (record, "defined_type_name"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["citation"],            value_or_none (record, "citation"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["metadata_reason"],     value_or_none (record, "metadata_reason"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["confidential_reason"], value_or_none (record, "confidential_reason"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["funding"],             value_or_none (record, "funding"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["size"],                value_or_none (record, "size"), XSD.integer)
            rdf.add (self.store, uri, rdf.DJHT["status"],              value_or_none (record, "status"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["version"],             value_or_none (record, "version"), XSD.integer)
            rdf.add (self.store, uri, rdf.DJHT["description"],         value_or_none (record, "description"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["figshare_url"],        value_or_none (record, "figshare_url"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["resource_doi"],        value_or_none (record, "resource_doi"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["resource_title"],      value_or_none (record, "resource_title"), XSD.string)
            rdf.add (self.store, uri, rdf.DJHT["has_linked_file"],     has_linked_file, XSD.boolean)
            rdf.add (self.store, uri, rdf.DJHT["is_metadata_record"],  is_metadata_record, XSD.boolean)
            rdf.add (self.store, uri, rdf.DJHT["is_confidential"],     is_confidential, XSD.boolean)
            rdf.add (self.store, uri, rdf.DJHT["is_public"],           is_public, XSD.boolean)
            rdf.add (self.store, uri, rdf.DJHT["is_latest"],           is_latest, XSD.boolean)
            rdf.add (self.store, uri, rdf.DJHT["is_editable"],         is_editable, XSD.boolean)

            if is_embargoed:
                rdf.add (self.store, uri, rdf.DJHT["embargo_until_date"], value_or_none (record, "embargo_date"), XSD.date)
                rdf.add (self.store, uri, rdf.DJHT["embargo_type"],    value_or_none (record, "embargo_type"), XSD.string)
                rdf.add (self.store, uri, rdf.DJHT["embargo_title"],   value_or_none (record, "embargo_title"), XSD.string)
                rdf.add (self.store, uri, rdf.DJHT["embargo_reason"],  value_or_none (record, "embargo_reason"), XSD.string)

            self.insert_item_list (uri, value_or (record, "references", []), "references")
            tags = [tag for tag in value_or (record, "tags", []) if not tag.startswith('Collection: ')]
            self.insert_item_list (uri, tags, "tags")
            self.insert_category_list (uri, value_or (record, "categories", []))
            self.insert_author_list (uri, value_or (record, "authors", []))
            self.insert_file_list (uri, value_or (record, "files", []))
            self.insert_funding_list (uri, value_or (record, "funding_list", []))
            self.insert_private_links_list (uri, value_or (record, "private_links", []))

            self.handle_custom_fields (record, uri, dataset_id, version, 'articles')

            ## Assign the dataset to the container
            container = self.container_uri (dataset_id, "dataset", account_id, account_uri)
            rdf.add (self.store, uri, rdf.DJHT["container"], container, datatype="uri")

            if "statistics" in record:
                stats = record["statistics"]
                self.insert_totals_statistics (stats["totals"], container)
            elif is_public and is_editable:
                logging.warning ("No statistics available for dataset %d.", dataset_id)

            if is_editable:
                self.store.add ((container, rdf.DJHT["draft"], uri))
            else:
                new_blank_node = rdf.blank_node ()
                self.store.add ((new_blank_node, RDF.type, RDF.List))
                self.store.add ((new_blank_node, RDF.first, uri))
                self.store.add ((new_blank_node, RDF.rest, RDF.nil))

                blank_node = self.last_list_node (container, "published_versions")
                if blank_node is None:
                    self.store.add    ((container, rdf.DJHT["published_versions"], new_blank_node))
                else:
                    self.store.remove ((blank_node, RDF.rest, RDF.nil))
                    self.store.add    ((blank_node, RDF.rest, new_blank_node))

            ## Insert the review state for the record.
            review = value_or_none (record, "review")
            if review:
                logging.info ("Adding review for dataset %s v%s.", uri, "0" if version is None else version)
                review_uri  = rdf.unique_node ("review")
                status      = value_or (review, "status", "pending")
                assigned_to = value_or_none (review, "assigned_to")
                assigned_to_uri = None
                status_uri  = None
                is_under_review = True
                if assigned_to is None:
                    status_uri = rdf.DJHT["ReviewUnassigned"]
                elif int(assigned_to) == 0:
                    status_uri = rdf.DJHT["ReviewUnassigned"]
                elif isinstance(status, str) and status == "pending":
                    status_uri = rdf.DJHT["ReviewAssigned"]
                    assigned_to_uri = self.record_uri ("Account", "id", assigned_to)
                elif isinstance(status, str):
                    status_uri = rdf.DJHT["Review" + status.capitalize()]
                    assigned_to_uri = self.record_uri ("Account", "id", assigned_to)
                    is_under_review = False
                else:
                    status_uri = rdf.DJHT["ReviewUnassigned"]

                rdf.add (self.store, uri, rdf.DJHT["is_under_review"], is_under_review, XSD.boolean)
                rdf.add (self.store, review_uri, RDF.type, rdf.DJHT["Review"], datatype="uri")
                rdf.add (self.store, review_uri, rdf.DJHT["dataset"], uri, datatype="uri")
                rdf.add (self.store, review_uri, rdf.DJHT["request_date"], value_or_none (review, "created_date"),  XSD.dateTime)
                rdf.add (self.store, review_uri, rdf.DJHT["modified_by_reviewer"], value_or_none (review, "modified_date"), XSD.dateTime)
                rdf.add (self.store, review_uri, rdf.DJHT["assigned_to"], assigned_to_uri, datatype="uri")
                rdf.add (self.store, review_uri, rdf.DJHT["status"], status_uri, datatype="uri")

            if is_latest:
                self.store.add ((container, rdf.DJHT["latest_published_version"], uri))
                self.add_container_doi (record, container)
                timeline = value_or_none (record, "timeline")
                rdf.add (self.store, container, rdf.DJHT["first_online_date"],
                         value_or_none (timeline, "firstOnline"), XSD.dateTime)

        return True

    def insert_institution_group (self, record):
        """Procedure to insert a institution group record."""

        uri        = rdf.unique_node ("institution_group")
        self.store.add ((uri, RDF.type, rdf.DJHT["InstitutionGroup"]))

        rdf.add (self.store, uri, rdf.DJHT["id"],                   value_or_none (record, "id"), datatype=XSD.integer)
        rdf.add (self.store, uri, rdf.DJHT["parent_id"],            value_or_none (record, "parent_id"), datatype=XSD.integer)
        ## resource_id is always empty.
        #rdf.add (self.store, uri, rdf.DJHT["resource_id"],          value_or_none (record, "resource_id"), datatype=XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["name"],                 value_or_none (record, "name"), datatype=XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["association_criteria"], value_or_none (record, "association_criteria"), datatype=XSD.string)
        rdf.add (self.store, uri, rdf.DJHT["is_featured"],          value_or_none (record, "is_featured"), datatype=XSD.boolean)

        return True

    def insert_static_triplets (self):
        """Procedure to insert triplets to augment the state graph."""

        self.store.add ((rdf.DJHT["DatasetContainer"],    RDFS.subClassOf, rdf.DJHT["Container"]))
        self.store.add ((rdf.DJHT["CollectionContainer"], RDFS.subClassOf, rdf.DJHT["Container"]))

        ## Review states from Figshare.
        self.store.add ((rdf.DJHT["ReviewApproved"],   RDF.type,   rdf.DJHT["ReviewType"]))
        self.store.add ((rdf.DJHT["ReviewApproved"],   RDFS.label, Literal("approved", datatype=XSD.string)))
        self.store.add ((rdf.DJHT["ReviewRejected"],   RDF.type,   rdf.DJHT["ReviewType"]))
        self.store.add ((rdf.DJHT["ReviewRejected"],   RDFS.label, Literal("rejected", datatype=XSD.string)))
        self.store.add ((rdf.DJHT["ReviewClosed"],     RDF.type,   rdf.DJHT["ReviewType"]))
        self.store.add ((rdf.DJHT["ReviewClosed"],     RDFS.label, Literal("closed", datatype=XSD.string)))

        ## We split "pending" into "assigned" and "unassigned" in Djehuty.
        self.store.add ((rdf.DJHT["ReviewAssigned"],   RDF.type,   rdf.DJHT["ReviewType"]))
        self.store.add ((rdf.DJHT["ReviewAssigned"],   RDFS.label, Literal("assigned", datatype=XSD.string)))
        self.store.add ((rdf.DJHT["ReviewUnassigned"], RDF.type,   rdf.DJHT["ReviewType"]))
        self.store.add ((rdf.DJHT["ReviewUnassigned"], RDFS.label, Literal("unassigned", datatype=XSD.string)))

        ## Log event types.
        self.store.add ((rdf.DJHT["LogEntryCite"],     RDF.type,   rdf.DJHT["LogEntryType"]))
        self.store.add ((rdf.DJHT["LogEntryCite"],     RDFS.label, Literal("cite", datatype=XSD.string)))
        self.store.add ((rdf.DJHT["LogEntryDownload"], RDF.type,   rdf.DJHT["LogEntryType"]))
        self.store.add ((rdf.DJHT["LogEntryDownload"], RDFS.label, Literal("download", datatype=XSD.string)))
        self.store.add ((rdf.DJHT["LogEntryGitDownload"], RDF.type,   rdf.DJHT["LogEntryType"]))
        self.store.add ((rdf.DJHT["LogEntryGitDownload"], RDFS.label, Literal("git_download", datatype=XSD.string)))
        self.store.add ((rdf.DJHT["LogEntryShare"],    RDF.type,   rdf.DJHT["LogEntryType"]))
        self.store.add ((rdf.DJHT["LogEntryShare"],    RDFS.label, Literal("share", datatype=XSD.string)))
        self.store.add ((rdf.DJHT["LogEntryView"],     RDF.type,   rdf.DJHT["LogEntryType"]))
        self.store.add ((rdf.DJHT["LogEntryView"],     RDFS.label, Literal("view", datatype=XSD.string)))
        self.store.add ((rdf.DJHT["LogEntryPrivateView"], RDF.type,   rdf.DJHT["LogEntryType"]))
        self.store.add ((rdf.DJHT["LogEntryPrivateView"], RDFS.label, Literal("private_view", datatype=XSD.string)))

        languages = self.__load_resource_file("languages.json")
        for language in languages:
            uri = rdf.unique_node ("language")
            self.store.add ((uri, RDF.type, rdf.DJHT["Language"]))
            self.store.add ((uri, RDFS.label,
                             Literal(language["name"], datatype=XSD.string)))
            self.store.add ((uri, rdf.DJHT["shortcode"],
                             Literal(language["shortcode"], datatype=XSD.string)))

        licenses = self.__load_resource_file ("licenses.json")
        for license_record in licenses:
            self.insert_license (None, license_record)

        categories = self.__load_resource_file ("root_categories.json")
        for category in categories:
            self.insert_category(category)

        return True



"""This module provides thread- and process locks for sensitive procedures."""

from enum import Enum
from threading import Lock
from djehuty.web.config import config

try:
    import uwsgi
except ModuleNotFoundError:
    pass

class LockTypes(Enum):
    """Enumeration of lock types."""
    FILE_LIST     = 1
    PRIVATE_LINKS = 2
    AUTHORS       = 3
    SUBMIT_DATASET = 4

class Locks:
    """This class implements multiple locks"""

    ## Only ever allow one instance of this class to exist,
    ## so that we prevent re-initializing locks.
    _instance = None
    def __new__ (cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__ (self):
        self.locks = {
            LockTypes.FILE_LIST: Lock(),
            LockTypes.PRIVATE_LINKS: Lock(),
            LockTypes.AUTHORS: Lock(),
            LockTypes.SUBMIT_DATASET: Lock(),
        }

    def lock (self, lock_type):
        """Lock critical section LOCK_TYPE."""

        if config.using_uwsgi:
            uwsgi.lock(lock_type.value)
        else:
            lock = self.locks.get(lock_type)
            lock.acquire(blocking=True, timeout=30)

    def unlock (self, lock_type):
        """Unlock critical section LOCK_TYPE."""
        if config.using_uwsgi:
            uwsgi.unlock(lock_type.value)
        else:
            lock = self.locks.get(lock_type)
            lock.release()



"""This module provides the functionality to send e-mails."""

import logging
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

class EmailInterface:
    """This class implements sending e-mail and e-mail templates."""

    def __init__ (self):
        self.smtp_server = None
        self.from_address = None
        self.smtp_username = None
        self.smtp_password = None
        self.smtp_port = 587
        self.do_starttls = True
        self.log = logging.getLogger(__name__)
        self.subject_prefix = None

    def is_properly_configured (self):
        """Procedure to bail early on a misconfigured instance of this class."""
        return (self.smtp_server is not None and
                self.from_address is not None)

    def send_email (self, recipient, subject, plaintext, html):
        """Procedure to send an email."""

        if not self.is_properly_configured ():
            self.log.error ("E-mail server not properly configured.")
            return False

        message = MIMEMultipart ("alternative")
        message["From"] = self.from_address
        message["To"] = recipient
        if self.subject_prefix:
            message["Subject"] = f"{self.subject_prefix} {subject}"
        else:
            message["Subject"] = subject

        message.attach (MIMEText (plaintext, "plain"))
        message.attach (MIMEText (html, "html"))

        try:
            connection = smtplib.SMTP (self.smtp_server, self.smtp_port, timeout=10)
        except (smtplib.SMTPConnectError, smtplib.SMTPException,
                smtplib.SMTPServerDisconnected, ConnectionRefusedError) as error:
            self.log.error ("Connecting to the e-mail server failed: %s", error)
            return False

        connection.ehlo()
        if self.do_starttls:
            connection.starttls ()
        elif self.smtp_username is not None and self.smtp_password is not None:
            self.log.error ("The e-mail interface hasn't been tested without STARTTLS.")
            self.log.error ("Please review the code before continuing.")
            return False
        else:
            self.log.warning ("STARTTLS is not enabled. Please review your e-mail settings.")

        if self.smtp_username is not None and self.smtp_password is not None:
            try:
                connection.login (self.smtp_username, self.smtp_password)
            except smtplib.SMTPAuthenticationError:
                self.log.error ("Wrong credentials for authenticating to the e-mail server.")
                return False
            except (smtplib.SMTPHeloError, smtplib.SMTPServerDisconnected,
                    smtplib.SMTPNotSupportedError, smtplib.SMTPException) as error:
                self.log.error ("Authenticating to the e-mail server failed: %s", error)
                return False

        try:
            connection.sendmail (self.from_address, recipient, message.as_string())
        except (smtplib.SMTPRecipientsRefused, smtplib.SMTPHeloError, smtplib.SMTPSenderRefused,
                smtplib.SMTPDataError, smtplib.SMTPNotSupportedError,
                smtplib.SMTPServerDisconnected) as error:
            self.log.error ("Sending e-mail failed: %s", error)
            return False

        connection.close()
        return True



"""
This module contains procedures to format a record from
djehuty.database to be backward-compatible with Figshare.
"""

from djehuty.utils import convenience as conv

def format_collaborator_record (record):
    """Record formatter for collaborators"""
    return {
        "uuid":           conv.value_or_none(record, "uuid"),
        "account_uuid":   conv.value_or_none (record, "account_uuid"),
        "first_name":     conv.value_or_none (record, "first_name"),
        "last_name":      conv.value_or_none(record, "last_name"),
        "email":          conv.value_or_none(record, "email"),
        "metadata_read":  conv.value_or_none (record, "metadata_read"),
        "metadata_edit":  conv.value_or_none(record, "metadata_edit"),
        "data_read":      conv.value_or_none(record, "data_read"),
        "data_edit":      conv.value_or_none(record, "data_edit"),
        "data_remove":    conv.value_or_none(record, "data_remove"),
        "is_supervisor":  conv.value_or_none (record, "is_supervisor"),
        "group_id":       conv.value_or_none (record, "group_id"),
        "group_name":     conv.value_or_none (record, "group_name"),
        "is_inferred":    conv.value_or(record, "is_inferred", False)
    }

def format_account_record (record):
    """Record formatter for accounts."""
    return {
        "id":             conv.value_or_none(record, "account_id"),
        "uuid":           conv.value_or_none(record, "uuid"),
        "first_name":     conv.value_or_none(record, "first_name"),
        "last_name":      conv.value_or_none(record, "last_name"),
        "full_name":      conv.value_or_none(record, "full_name"),
        "is_active":      bool(conv.value_or_none(record, "active")),
        "is_public":      bool(conv.value_or_none(record, "public")),
        "job_title":      conv.value_or_none(record, "job_title"),
        "orcid_id":       conv.value_or (record, "orcid_id", ""),
    }

def format_account_details_record (record):
    """Record formatter for accounts."""
    return {
        "id":             conv.value_or_none(record, "account_id"),
        "uuid":           conv.value_or_none(record, "uuid"),
        "first_name":     conv.value_or_none(record, "first_name"),
        "last_name":      conv.value_or_none(record, "last_name"),
        "full_name":      conv.value_or_none(record, "full_name"),
        "email":          conv.value_or_none(record, "email"),
        "is_active":      bool(conv.value_or_none(record, "active")),
        "is_public":      bool(conv.value_or_none(record, "public")),
        "job_title":      conv.value_or_none(record, "job_title"),
        "orcid_id":       conv.value_or (record, "orcid_id", ""),
    }

def collection_urls (record):
    """Returns generated variants of public-facing collection URLs when possible."""
    if "base_url" in record:
        version = ""
        if "version" in record:
            version = f"/{record['version']}"

        return {
            "url":              f"{record['base_url']}/v2/collections/{record['container_uuid']}",
            "url_private_api":  f"{record['base_url']}/v2/account/collections/{record['container_uuid']}",
            "url_public_api":   f"{record['base_url']}/v2/collections/{record['container_uuid']}",
            "url_private_html": f"{record['base_url']}/my/collections/{record['container_uuid']}/edit",
            "url_public_html":  f"{record['base_url']}/collections/{record['container_uuid']}{version}"
        }

    return {
        "url":              conv.value_or_none(record, "url"),
        "url_private_api":  conv.value_or_none(record, "url_private_api"),
        "url_public_api":   conv.value_or_none(record, "url_public_api"),
        "url_private_html": conv.value_or_none(record, "url_private_html"),
        "url_public_html":  conv.value_or_none(record, "url_public_html")
    }

def dataset_urls (record):
    """Returns generated variants of public-facing dataset URLs when possible."""

    if "base_url" in record:
        version = ""
        if "version" in record:
            version = f"/{record['version']}"
        return {
            "url":              f"{record['base_url']}/v2/articles/{record['container_uuid']}",
            "url_private_api":  f"{record['base_url']}/v2/account/articles/{record['container_uuid']}",
            "url_public_api":   f"{record['base_url']}/v2/articles/{record['container_uuid']}",
            "url_private_html": f"{record['base_url']}/my/datasets/{record['container_uuid']}/edit",
            "url_public_html":  f"{record['base_url']}/datasets/{record['container_uuid']}{version}"
        }

    return {
        "url":              conv.value_or_none(record, "url"),
        "url_private_api":  conv.value_or_none(record, "url_private_api"),
        "url_public_api":   conv.value_or_none(record, "url_public_api"),
        "url_private_html": conv.value_or_none(record, "url_private_html"),
        "url_public_html":  conv.value_or_none(record, "url_public_html")
    }

def format_codemeta_author_record (record, base_url):
    """Record formatter for an author as used in CodeMeta."""
    author = {
        "@type": "Person",
        "givenName": conv.value_or_none (record, "first_name"),
        "familyName": conv.value_or_none (record, "last_name"),
        "name": conv.value_or_none (record, "full_name"),
        "email": None,
    }

    orcid = conv.value_or_none (record, "orcid_id")
    if orcid is not None:
        author["@id"] = f"https://orcid.org/{orcid}"
    else:
        author["@id"] = f"{base_url}/authors/{record['uuid']}"

    return author

def format_codemeta_record (record, git_url, tags, authors, has_files, base_url):
    """Record formatter for the CodeMeta format."""

    title = conv.value_or_none (record, "git_repository_name")
    if title is None or title == "":
        title = conv.value_or_none (record, "title")

    if bool(conv.value_or (record, "is_embargoed", False)):
        return {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "@type": "SoftwareSourceCode",
            "name": title,
            "dateCreated": conv.value_or_none(record, "created_date"),
            "datePublished": conv.value_or_none(record, "published_date"),
            "embargoDate": conv.value_or_none(record, "embargo_until_date")
        }

    # Strip the HTML tags from the description.
    description = conv.value_or_none (record, "description")
    if description:
        description = conv.html_to_plaintext (description, True)

    output = {
        "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
        "@type": "SoftwareSourceCode",
        "name": title,
        "license": conv.value_or_none (record, "license_spdx"),
        "dateCreated": conv.value_or_none(record, "created_date"),
        "datePublished": conv.value_or_none(record, "published_date"),
        "dateModified": conv.value_or_none(record, "modified_date"),
        "identifier": conv.value_or_none(record, "doi"),
        "description": [conv.value_or (record, "title", ""), description],
        "keywords": list (map (format_tag_record, tags)),
        "author": list (map (lambda author: format_codemeta_author_record (author, base_url), authors))
    }

    resource_doi = conv.value_or_none (record, "resource_doi")
    if resource_doi is not None:
        output["referencePublication"] = [{
            "@type": "ScholarlyArticle",
            "url": f"https://doi.org/{resource_doi}",
            "name": conv.value_or_none (record, "resource_title"),
            "@id": resource_doi
        }]
    else:
        output["referencePublication"] = []

    download_url = []
    if has_files and "version" in record and "container_uuid" in record:
        download_url = [f"{base_url}/ndownloader/items/{record['container_uuid']}/versions/{record['version']}"]

    if git_url:
        download_url.append (f"{git_url}/zip")
        output["codeRepository"] = git_url

    if download_url:
        output["downloadUrl"] = download_url

    if "version" in record:
        output["version"] = record["version"]

    return output

def format_rocrate_file_record (record):
    """Record formatter for RO-Crate file objects."""

    return {
        "@id": record["uuid"],
        "@type": "File",
        "name": conv.value_or_none (record, "name"),
        "contentSize": f"{conv.value_or_none(record, 'size')}",
        "contentUrl": file_download_url (record)
    }

def format_rocrate_record (base_url, site_name, record, ror_url, tags,
                           authors, files, git_url=None):
    """Record formatter for the RO-Crate format."""

    # Strip the HTML tags from the description.
    description = conv.value_or_none (record, "description")
    if description:
        description = conv.html_to_plaintext (description, True)

    ro_file_ids = []
    for file in files:
        file["base_url"] = base_url
        ro_file_ids.append({ "@id": file["uuid"] })

    file_records = list (map (format_rocrate_file_record, files))
    publisher_records = []
    git_record = []
    ro_crate_meta_record = {
        "@id": "ro-crate-metadata.json",
        "@type": "CreativeWork",
        "conformsTo": { "@id": "https://w3id.org/ro/crate/1.1" },
        "about": { "@id": "./" }
    }
    dataset_record = {
        "@id": "./",
        "@type": "Dataset",
        "identifier": f"https://doi.org/{record['doi']}",
        "datePublished": conv.value_or_none(record, "published_date"),
        "name": conv.value_or_none (record, "title"),
        "description": description,
        "keywords": list (map (format_tag_record, tags)),
        "license": { "@id": conv.value_or_none (record, "license_spdx") },
        "author": list (map (lambda author: format_codemeta_author_record (author, base_url), authors)),
        "hasPart": list (map (lambda item: { "@id": item["@id"] }, file_records))
    }

    if git_url:
        title = conv.value_or_none (record, "git_repository_name")
        if title is None or title == "":
            title = conv.value_or_none (record, "title")

        git_record = [{
            "@id": record["git_uuid"],
            "@type": "SoftwareSourceCode",
            "name": title,
            "codeRepository": git_url,
        }]
        dataset_record["hasPart"].append({ "@id": record["git_uuid"] })

    if ror_url:
        publisher_records = [{
            "@id": ror_url,
            "@type": "Organization",
            "name": site_name,
            "url": base_url
        }]
        dataset_record["publisher"] = { "@id": ror_url }

    return {
        "@context": "https://w3id.org/ro/crate/1.1/context",
        "@graph": [ro_crate_meta_record, dataset_record] + git_record + publisher_records + file_records
    }

def format_dataset_record (record):
    """Record formatter for datasets."""

    urls = dataset_urls (record)
    if (bool(conv.value_or (record, "is_embargoed", False)) and
        conv.value_or (record, "embargo_type", "") == "article"):
        return {
            "uuid":           conv.value_or_none(record, "container_uuid"),
            "title":          conv.value_or_none(record, "title"),
            "embargo_date":   conv.value_or_none(record, "embargo_until_date"),
            "embargo_type":   conv.value_or_none(record, "embargo_type"),
            "embargo_title":  conv.value_or(record, "embargo_title", ""),
            "embargo_reason": conv.value_or(record, "embargo_reason", ""),
            "published_date": conv.value_or_none(record, "published_date"),
            "url_public_html": conv.value_or_none(urls, "url_public_html"),
            "timeline": {
                "posted":     conv.value_or_none(record, "timeline_posted"),
                "firstOnline": conv.value_or_none(record, "timeline_first_online"),
                "revision":   conv.value_or_none(record, "timeline_revision"),
            }
        }

    return {
        "id":                      conv.value_or_none(record, "dataset_id"),
        "uuid":                    conv.value_or_none(record, "container_uuid"),
        "title":                   conv.value_or_none(record, "title"),
        "doi":                     conv.value_or_none(record, "doi"),
        "handle":                  conv.value_or_none(record, "handle"),
        "url":                     conv.value_or_none(urls, "url"),
        "published_date":          conv.value_or_none(record, "published_date"),
        "thumb":                   conv.value_or_none(record, "thumb"),
        "defined_type":            conv.value_or_none(record, "defined_type"),
        "defined_type_name":       conv.value_or_none(record, "defined_type_name"),
        "group_id":                conv.value_or_none(record, "group_id"),
        "url_private_api":         conv.value_or_none(urls, "url_private_api"),
        "url_public_api":          conv.value_or_none(urls, "url_public_api"),
        "url_private_html":        conv.value_or_none(urls, "url_private_html"),
        "url_public_html":         conv.value_or_none(urls, "url_public_html"),
        "timeline": {
            "posted":              conv.value_or_none(record, "timeline_posted"),
            "firstOnline":         conv.value_or_none(record, "timeline_first_online"),
            "revision":            conv.value_or_none(record, "timeline_revision"),
        },
        "resource_title":          conv.value_or_none(record, "resource_title"),
        "resource_doi":            conv.value_or_none(record, "resource_doi")
    }

def format_author_record_v3 (record):
    """Record formatter for the v3 authors API output."""
    email = None
    if conv.value_or (record, "is_editable", False):
        email = conv.value_or_none (record, "email")

    return {
        "uuid":        conv.value_or_none (record, "uuid"),
        "first_name":  conv.value_or_none(record, "first_name"),
        "last_name":   conv.value_or_none(record, "last_name"),
        "full_name":   conv.value_or_none(record, "full_name"),
        "email":       email,
        "orcid":       conv.value_or (record, "orcid_id", None),
        "is_editable": conv.value_or (record, "is_editable", False)
    }

def format_author_record (record):
    """Record formatter for authors."""
    return {
      "id":        conv.value_or_none(record, "id"),
      "uuid":      conv.value_or_none(record, "uuid"),
      "full_name": conv.value_or_none(record, "full_name"),
      "is_active": bool(conv.value_or_none(record, "is_active")),
      "url_name":  conv.value_or_none(record, "url_name"),
      "orcid_id":  conv.value_or(record, "orcid_id", "")
    }

def format_author_details_record (record):
    """Detailed record formatter for authors."""
    return {
      "first_name":     conv.value_or_none(record, "first_name"),
      "full_name":      conv.value_or_none(record, "full_name"),
      "group_id":       conv.value_or_none(record, "group_id"),
      "id":             conv.value_or_none(record, "id"),
      "uuid":           conv.value_or_none(record, "uuid"),
      "institution_id": conv.value_or_none(record, "institution_id"),
      "is_active":      bool(conv.value_or_none(record, "is_active")),
      "is_public":      bool(conv.value_or_none(record, "is_public")),
      "job_title":      conv.value_or_none(record, "job_title"),
      "last_name":      conv.value_or_none(record, "last_name"),
      "orcid_id":       conv.value_or (record, "orcid_id", ""),
      "url_name":       conv.value_or_none(record, "url_name")
    }

def file_download_url (record):
    """Returns a generated download_url or its default."""

    if "base_url" in record and not conv.value_or(record, "is_link_only", False):
        return f"{record['base_url']}/file/{record['container_uuid']}/{record['uuid']}"

    return conv.value_or_none(record, "download_url")

def format_file_for_dataset_record (record):
    """Record formatter for files."""
    download_url = file_download_url (record)
    output = {
      "id":            conv.value_or_none(record, "id"),
      "uuid":          conv.value_or_none(record, "uuid"),
      "name":          conv.value_or_none(record, "name"),
      "size":          conv.value_or_none(record, "size"),
      "is_link_only":  bool(conv.value_or_none(record, "is_link_only")),
      "is_incomplete": bool(conv.value_or_none(record, "is_incomplete")),
      "download_url":  download_url,
      "supplied_md5":  conv.value_or_none(record, "supplied_md5"),
      "computed_md5":  conv.value_or_none(record, "computed_md5"),
    }
    if "handle" in record:
        output["handle"] = record["handle"]

    return output

def format_file_details_record (record):
    """Detailed record formatter for files."""
    download_url = file_download_url (record)
    return {
      "status":        conv.value_or_none(record, "status"),
      "viewer_type":   conv.value_or_none(record, "viewer_type"),
      "preview_state": conv.value_or_none(record, "preview_state"),
      "upload_url":    conv.value_or_none(record, "upload_url"),
      "upload_token":  conv.value_or_none(record, "upload_token"),
      "uuid":          conv.value_or_none(record, "uuid"),
      "id":            conv.value_or_none(record, "id"),
      "name":          conv.value_or_none(record, "name"),
      "size":          conv.value_or_none(record, "size"),
      "is_link_only":  conv.value_or_none(record, "is_link_only"),
      "download_url":  download_url,
      "supplied_md5":  conv.value_or_none(record, "supplied_md5"),
      "computed_md5":  conv.value_or_none(record, "computed_md5")
    }

def format_custom_field_record (record):
    """Record formatter for custom fields."""
    value = conv.value_or_none (record, "value")
    name  = conv.value_or_none (record, "name")
    if name is not None and name == "Data Link":
        value = [value]

    return { "name": name, "value": value }

def format_category_record (record):
    """Record formatter for categories."""
    return {
        "id":          conv.value_or_none(record, "id"),
        "uuid":        conv.value_or_none(record, "uuid"),
        "title":       conv.value_or_none(record, "title"),
        "parent_id":   conv.value_or_none(record, "parent_id"),
        "parent_uuid": conv.value_or_none(record, "parent_uuid"),
        "path":        conv.value_or(record, "path", ""),
        "source_id":   conv.value_or_none(record, "source_id"),
        "taxonomy_id": conv.value_or_none(record, "taxonomy_id"),
    }

def format_tag_record (record):
    """Record formatter for tags."""
    return conv.value_or_none(record, "tag")

def format_reference_record (record):
    """Record formatter for references."""
    return conv.value_or_none(record, "url")

def format_license_record (record):
    """Record formatter for licenses."""
    return {
        "value":         conv.value_or_none(record, "id"),
        "name":          conv.value_or_none(record, "name"),
        "url":           conv.value_or_none(record, "url"),
        # Extension for Djehuty.
        "type":          conv.value_or_none(record, "type"),
    }

def format_dataset_details_record (dataset, authors, files, custom_fields,
                                   tags, categories, funding, references,
                                   is_private=False):
    """Detailed record formatter for datasets."""

    is_embargoed = bool(conv.value_or (dataset, "is_embargoed", False))
    is_restricted = bool(conv.value_or (dataset, "is_restricted", False))
    if ((is_embargoed or is_restricted) and
        conv.value_or (dataset, "embargo_type", "") == "article" and
        not is_private):
        return {
            "embargo_date":      conv.value_or_none(dataset, "embargo_until_date"),
            "embargo_type":      conv.value_or_none(dataset, "embargo_type"),
            "embargo_title":     conv.value_or(dataset, "embargo_title", ""),
            "embargo_reason":    conv.value_or(dataset, "embargo_reason", ""),
        }

    embargo_option = None
    if is_restricted:
        embargo_option = { "id": 1000, "type": "restricted_access" }

    urls = dataset_urls (dataset)
    if "base_url" in dataset:
        for item in files:
            item['base_url'] = dataset['base_url']

    files_shown = []
    if not (is_embargoed or is_restricted):
        files_shown = list (map (format_file_for_dataset_record, files))

    return {
        "files":             files_shown,
        "custom_fields":     list (map (format_custom_field_record, custom_fields)),
        "authors":           list (map (format_author_record, authors)),
        "figshare_url":      conv.value_or_none(dataset, "figshare_url"),
        "description":       conv.value_or_none(dataset, "description"),
        "funding":           conv.value_or_none(dataset, "funding"),
        "funding_list":      list (map (format_funding_record, funding)),
        "version":           conv.value_or_none(dataset, "version"),
        "status":            conv.value_or_none(dataset, "status"),
        "size":              conv.value_or_none(dataset, "size"),
        "created_date":      conv.value_or_none(dataset, "created_date"),
        "modified_date":     conv.value_or_none(dataset, "modified_date"),
        "is_public":         bool(conv.value_or_none(dataset, "is_public")),
        "is_confidential":   bool(conv.value_or_none(dataset, "is_confidential")),
        "is_metadata_record": bool(conv.value_or_none(dataset, "is_metadata_record")),
        "confidential_reason": conv.value_or(dataset, "confidential_reason", ""),
        "metadata_reason":   conv.value_or(dataset, "metadata_reason", ""),
        "license": {
            "value":         conv.value_or_none(dataset, "license_id"),
            "name":          conv.value_or_none(dataset, "license_name"),
            "url":           conv.value_or_none(dataset, "license_url"),
        },
        "tags":              list (map (format_tag_record, tags)),
        "categories":        list (map (format_category_record, categories)),
        "references":        list (map (format_reference_record, references)),
        "has_linked_file":   bool(conv.value_or_none(dataset, "has_linked_file")),
        "citation":          conv.value_or_none(dataset, "citation"),
#        "is_active":         conv.value_or_none(dataset, "is_active"),
        "is_embargoed":      bool(is_embargoed or is_restricted),
        "embargo_date":      conv.value_or_none(dataset, "embargo_until_date"),
        "embargo_type":      conv.value_or_none(dataset, "embargo_type"),
        "embargo_title":     conv.value_or(dataset, "embargo_title", ""),
        "embargo_reason":    conv.value_or(dataset, "embargo_reason", ""),
        "embargo_options":   [embargo_option] if embargo_option is not None else [],
        "id":                conv.value_or_none(dataset, "dataset_id"),
        "uuid":              conv.value_or_none(dataset, "container_uuid"),
        "title":             conv.value_or_none(dataset, "title"),
        "doi":               conv.value_or_none(dataset, "doi"),
        "handle":            conv.value_or(dataset, "handle", ""),
        "url":               conv.value_or_none(urls, "url"),
        "published_date":    conv.value_or_none(dataset, "published_date"),
        "thumb":             conv.value_or(dataset, "thumb", ""),
        "defined_type":      conv.value_or_none(dataset, "defined_type"),
        "defined_type_name": conv.value_or_none(dataset, "defined_type_name"),
        "group_id":          conv.value_or_none(dataset, "group_id"),
        "url_private_api":   conv.value_or_none(urls, "url_private_api"),
        "url_public_api":    conv.value_or_none(urls, "url_public_api"),
        "url_private_html":  conv.value_or_none(urls, "url_private_html"),
        "url_public_html":   conv.value_or_none(urls, "url_public_html"),
        "timeline": {
            "posted":        conv.value_or_none(dataset, "timeline_posted"),
            "revision":      conv.value_or_none(dataset, "timeline_revision"),
            "submission":    conv.value_or_none(dataset, "timeline_submission"),
            "firstOnline":   conv.value_or_none(dataset, "timeline_first_online"),
            "publisherPublication": conv.value_or_none(dataset, "timeline_publisher_publication"),
        },
        "resource_title":    conv.value_or(dataset, "resource_title", ""),
        "resource_doi":      conv.value_or(dataset, "resource_doi", ""),
        "agreed_to_deposit_agreement": bool(conv.value_or_none (dataset, "agreed_to_deposit_agreement")),
        "agreed_to_publish": bool(conv.value_or_none(dataset, "agreed_to_publish"))
    }

def format_dataset_embargo_record (dataset):
    """Record formatter for embargos."""

    is_embargoed = bool(conv.value_or (dataset, "is_embargoed", False))
    is_restricted = bool(conv.value_or (dataset, "is_restricted", False))

    embargo_options = []
    if is_restricted:
        embargo_options = [{ "id": 1000, "type": "restricted_access" }]

    return {
        "is_embargoed":      bool(is_embargoed or is_restricted),
        "embargo_date":      conv.value_or_none(dataset, "embargo_until_date"),
        "embargo_type":      conv.value_or(dataset, "embargo_type", "file"),
        "embargo_title":     conv.value_or(dataset, "embargo_title", ""),
        "embargo_reason":    conv.value_or(dataset, "embargo_reason", ""),

        # Embargo options are irrelevant outside of Figshare.
        "embargo_options":   embargo_options,
    }

def format_dataset_confidentiality_record (dataset):
    """Record formatter for confidentiality."""
    return {
        "is_confidential":   bool(conv.value_or_none(dataset, "is_confidential")),
        "reason": conv.value_or(dataset, "confidential_reason", ""),
    }

def format_collection_record (record):
    """Record formatter for collections."""
    urls = collection_urls (record)
    return {
        "id":                conv.value_or_none(record, "collection_id"),
        "uuid":              conv.value_or_none(record, "container_uuid"),
        "title":             conv.value_or_none(record, "title"),
        "doi":               conv.value_or_none(record, "doi"),
        "handle":            conv.value_or(record, "handle", ""),
        "url":               conv.value_or_none(urls, "url"),
        "timeline": {
            "posted":        conv.value_or_none(record, "timeline_posted"),
            "submission":    conv.value_or_none(record, "timeline_submission"),
            "revision":      conv.value_or_none(record, "timeline_revision"),
            "firstOnline":   conv.value_or_none(record, "timeline_first_online"),
            "publisherPublication": conv.value_or_none(record, "timeline_publisher_publication"),
        },
        "published_date":    conv.value_or_none(record, "published_date"),
    }

def format_collection_details_record (collection, funding, categories,
                                      references, tags, authors, custom_fields,
                                      datasets_count):
    """Detailed record formatter for collections."""
    urls = collection_urls (collection)
    return {
        "version":           conv.value_or_none(collection, "version"),
        "resource_id":       conv.value_or(collection, "resource_id", ""),
        "resource_doi":      conv.value_or(collection, "resource_doi", ""),
        "resource_title":    conv.value_or(collection, "resource_title", ""),
        "resource_link":     conv.value_or(collection, "resource_link", ""),
        "resource_version":  conv.value_or_none(collection, "resource_version"),
        "description":       conv.value_or_none(collection, "description"),
        "categories":        list (map (format_category_record, categories)),
        "references":        list (map (format_reference_record, references)),
        "tags":              list (map (format_tag_record, tags)),
        "authors":           list (map (format_author_record, authors)),
        "funding":           list (map (format_funding_record, funding)),
        "institution_id":    conv.value_or_none(collection, "institution_id"),
        "group_id":          conv.value_or_none(collection, "group_id"),
        "group_resource_id": conv.value_or_none(collection, "group_resource_id"),
        "articles_count":    0 if datasets_count is None else datasets_count,
        "public":            bool(conv.value_or_none(collection, "is_public")),
        "custom_fields":     list (map (format_custom_field_record, custom_fields)),
        "citation":          conv.value_or_none(collection, "citation"),
        "created_date":      conv.value_or_none(collection, "created_date"),
        "modified_date":     conv.value_or_none(collection, "modified_date"),
        "id":                conv.value_or_none(collection, "collection_id"),
        "uuid":              conv.value_or_none(collection, "container_uuid"),
        "title":             conv.value_or_none(collection, "title"),
        "doi":               conv.value_or_none(collection, "doi"),
        "handle":            conv.value_or(collection, "handle", ""),
        "url":               conv.value_or_none(urls, "url"),
        "published_date":    conv.value_or_none(collection, "published_date"),
        "timeline": {
            "posted":               conv.value_or_none(collection, "timeline_posted"),
            "firstOnline":          conv.value_or_none(collection, "timeline_first_online"),
            "revision":             conv.value_or_none(collection, "timeline_revision"),
            "submission":           conv.value_or_none(collection, "timeline_submission"),
            "publisherPublication": conv.value_or_none(collection, "timeline_publisher_publication")
        }
    }

def format_funding_record (record):
    """Record formatter for funding."""
    return {
        "id":                conv.value_or_none(record, "id"),
        "uuid":              conv.value_or_none(record, "uuid"),
        "title":             conv.value_or_none(record, "title"),
        "grant_code":        conv.value_or_none(record, "grant_code"),
        "funder_name":       conv.value_or_none(record, "funder_name"),
        "is_user_defined":   conv.value_or_none(record, "is_user_defined"),
        "url":               conv.value_or_none(record, "url")
    }

def format_collection_version_record (record):
    """Record formatter for versions of collections."""
    version = conv.value_or_none (record, "version")
    urls    = collection_urls (record)
    return {
        "version": version,
        "url": f"{urls['url_public_api']}/versions/{version}"
    }

def format_dataset_version_record (record):
    """Record formatter for versions of datasets."""
    version = conv.value_or_none (record, "version")
    urls    = dataset_urls (record)
    return {
        "version": version,
        "url": f"{urls['url_public_api']}/versions/{version}"
    }

def format_private_links_record (record):
    """Record formatter for private links."""
    return {
        "id":           conv.value_or_none(record, "id_string"),
        "is_active":    bool(conv.value_or_none(record, "is_active")),
        "expires_date": conv.value_or_none(record, "expires_date")
    }

def format_group_record (record):
    """Record formatter for groups."""
    return {
      "id":            conv.value_or_none(record, "id"),
      "parent_id":     conv.value_or_none(record, "parent_id"),
      "name":          conv.value_or_none(record, "name"),
      "association":   conv.value_or_none(record, "association"),
      "is_featured":   conv.value_or_none(record, "is_featured")
    }

def format_review_record (record):
    """Record formatter for reviews."""
    return {
        "uuid":                  conv.value_or_none (record, "uuid"),
        "container_uuid":        conv.value_or_none (record, "container_uuid"),
        "dataset_title":         conv.value_or_none (record, "dataset_title"),
        "dataset_uuid":          conv.value_or_none (record, "dataset_uuid"),
        "dataset_version":       conv.value_or_none (record, "dataset_version"),
        "group_name":            conv.value_or_none (record, "group_name"),
        "has_published_version": conv.value_or_none (record, "has_published_version"),
        "last_seen_by_reviewer": conv.value_or_none (record, "last_seen_by_reviewer"),
        "modified_date":         conv.value_or_none (record, "modified_date"),
        "published_date":        conv.value_or_none (record, "published_date"),
        "request_date":          conv.value_or_none (record, "request_date"),
        "reviewer_email":        conv.value_or_none (record, "reviewer_email"),
        "reviewer_first_name":   conv.value_or_none (record, "reviewer_first_name"),
        "reviewer_last_name":    conv.value_or_none (record, "reviewer_last_name"),
        "status":                conv.value_or_none (record, "status"),
        "submitter_email":       conv.value_or_none (record, "submitter_email"),
        "submitter_first_name":  conv.value_or_none (record, "submitter_first_name"),
        "submitter_last_name":   conv.value_or_none (record, "submitter_last_name")
    }



"""
This module contains the entry point for the program.
"""

import argparse
import signal
import sys
import logging
import os
import importlib.metadata

import djehuty.backup.ui as backup_ui
import djehuty.web.ui as web_ui

def show_version ():
    """Show the program's version."""

    version = importlib.metadata.version (__package__ or __name__)
    print(f"This is djehuty v{version}")
    sys.exit(0)

def show_help ():
    """Show a GNU-style help message."""

    print("""This is djehuty.\n
Available subcommands and options:

  backup:
    --help               -h Show a help message.
    --stats-auth=ARG     -a Username/password for the statistics endpoint.
    --token=ARG          -t The API token to use.
    --account-id=ARG     -i The account ID to backup.
    --api-url            -u The base URL for accessing the API. Defaults to
                            'https://api.figshare.com'.

  web:
    --help               -h Show a help message.
    --config-file=ARG    -c Load configuration from a file.
    --initialize         -i Populate the RDF store with default triples.
    --extract-transactions-from-log=ARG
                         -e Extract transactions from the log file.  It does
                            not extract delayed transactions, so it matches
                            what would have been applied to the database
                            normally.  The optional argument is a datetime to
                            start from, in the format of YYYY-MM-DD HH:MM:SS.
    --extract-delayed-transactions-from-log=ARG
                         -d Similar to -e, except this option only extracts
                            delayed transactions from the log file.
    --apply-transactions=ARG
                         -A Apply transactions extracted by the -e option above.
                            The optional argument is the path to the folder that
                            contains the SPARQL transaction files.  By default
                            it will look in the pre-configured directory.
    --full-rdf-export    -f Exports all metadata as RDF triples in N3 format.
                            This feature is in BETA state. Do not rely on it
                            for a full back-up.
    --public-rdf-export  -p Exports all publically available metadata as RDF
                            triples in N3 format.  This feature is in BETA
                            state. Do not reply on it for an accurate
                            representation of publically available metadata.

  Global options:
  --help                 -h  Show this message.
  --version              -v  Show versioning information.\n""")
    sys.exit(0)

def sigint_handler (sig, frame):  # pylint: disable=unused-argument
    """Signal handler for SIGINT and SIGTERM."""
    logger = logging.getLogger(__name__)
    logger.info ("Received shutdown signal.  Goodbye!")
    sys.exit(0)

def main_inner ():
    """The main entry point of the program."""

    signal.signal(signal.SIGINT, sigint_handler)
    signal.signal(signal.SIGTERM, sigint_handler)

    logging.basicConfig(format='[%(levelname)s] %(asctime)s - %(name)s: %(message)s',
                        level=logging.INFO)

    ## COMMAND-LINE ARGUMENTS
    ## ------------------------------------------------------------------------
    parser = argparse.ArgumentParser(
        usage    = '\n  %(prog)s [backup|web] ...',
        prog     = 'djehuty',
        add_help = False)

    subparsers = parser.add_subparsers(dest='command', help='sub-command help')

    ### BACKUP SUBCOMMAND
    ### -----------------------------------------------------------------------
    backup_parser = subparsers.add_parser('backup', help="Options for the 'backup' subcommand.")
    backup_parser.add_argument('--token',       '-t', type=str, default='')
    backup_parser.add_argument('--stats-auth',  '-a', type=str, default='')
    backup_parser.add_argument('--account-id',  '-i', type=str, default=None)
    backup_parser.add_argument('--api-url',     '-u', type=str, default=None)

    ### WEB SUBCOMMAND
    ### -----------------------------------------------------------------------
    web_parser = subparsers.add_parser('web', help="Options for the 'web' subcommand.")
    web_parser.add_argument('--config-file','-c', type=str, default=None)
    web_parser.add_argument('--initialize', '-i', action='store_true')
    web_parser.add_argument('--extract-transactions-from-log', '-e', nargs='?',
                            const='', default=None)
    web_parser.add_argument('--extract-delayed-transactions-from-log', '-d', nargs='?',
                            const='', default=None)
    web_parser.add_argument('--apply-transactions', '-A', nargs='?',
                            const='', default=None)
    web_parser.add_argument('--full-rdf-export', '-f', action='store_true')
    web_parser.add_argument('--public-rdf-export', '-p', action='store_true')

    ### GLOBAL ARGUMENTS
    ### -----------------------------------------------------------------------
    parser.add_argument('--help',    '-h', action='store_true')
    parser.add_argument('--version', '-v', action='store_true')

    # When using PyInstaller and Nuitka, argv[0] seems to get duplicated.
    # In the case of Nuitka, relative paths are converted to absolute paths.
    # This bit de-duplicates argv[0] in these cases.
    try:
        if os.path.abspath(sys.argv[0]) == os.path.abspath(sys.argv[1]):
            sys.argv = sys.argv[1:]
    except IndexError:
        pass

    args = parser.parse_args()
    if args.help:
        show_help()
    if args.version:
        show_version()

    if args.command == "backup":
        if "" in (args.token, args.stats_auth):
            print ("The 'backup' command requires multiple arguments.")
            print ("Try --help for usage options.")
        else:
            backup_ui.main (args.token, args.stats_auth, args.account_id,
                            args.api_url)

    elif args.command == "web":
        web_ui.main (args.config_file, True, args.initialize,
                     args.extract_transactions_from_log,
                     args.extract_delayed_transactions_from_log,
                     args.apply_transactions, args.full_rdf_export,
                     args.public_rdf_export)

    elif len(sys.argv) == 1:
        print("Try --help for usage options.")

def main ():
    """Wrapper to catch KeyboardInterrupts for main_inner."""
    try:
        main_inner()
    except KeyboardInterrupt:
        sigint_handler (None, None)

    sys.exit(0)



"""
This module provides convenience functions for handling RDF.
"""

import re
import uuid
from rdflib import Literal, Namespace, URIRef, XSD

BLANK = Namespace("blank:")
ROW   = Namespace("djehuty:")
DJHT  = Namespace("https://ontologies.data.4tu.nl/djehuty/0.0.1/")

## Pre-compiled patterns for determining the SPARQL query type.
COMMENTS_PATTERN = re.compile(r"(^|\n)\s*#.*?\n")
PREFIX_PATTERN   = re.compile(
    r"((?P<base>(\s*BASE\s*<.*?>)\s*)|(?P<prefixes>(\s*PREFIX\s+.+:\s*<.*?>)\s*))*"
)
QUERY_PATTERN    = re.compile(
    r"(?P<queryType>(CONSTRUCT|SELECT|ASK|DESCRIBE|INSERT|DELETE|CREATE|CLEAR|DROP|LOAD|COPY|MOVE|ADD))",
    re.VERBOSE | re.IGNORECASE,
)

def query_type (query):
    """
    Returns two values. The first value is 'update' for state-modifying
    queries, 'gather' for read-only queries, None when parsing failed.
    The second value is the more precise query type, like SELECT, INSERT
    or DELETE, or None when parsing failed.

    The implementation is based on SPARQLWrapper's implementation.
    """

    query_wo_comments = re.sub(COMMENTS_PATTERN, "\n\n", query)
    simplified_query  = re.sub(PREFIX_PATTERN, "", query_wo_comments.strip())
    try:
        sparql_type = str(QUERY_PATTERN.search(simplified_query).group("queryType").upper())
        if sparql_type in [ "SELECT", "CONSTRUCT", "ASK" ]:
            return "gather", sparql_type
        if sparql_type in [ "INSERT", "DELETE", "CLEAR", "DROP", "LOAD"]:
            return "update", sparql_type
    except AttributeError:
        pass

    return None, None

def add (graph, subject, predicate, value, datatype=None):
    """Adds the triplet SUBJECT PREDICATE VALUE if VALUE is set."""
    if value is not None:
        if isinstance (value, str) and value == "" and datatype == XSD.integer:
            return None

        if datatype in ("url", "uri"):
            graph.add ((subject, predicate, URIRef(value)))
        else:
            graph.add((subject, predicate, Literal(value, datatype=datatype)))

    return None

def urify_value (value):
    """Returns a string including smaller-than and greater-than brackets."""
    if isinstance(value, str) and value.startswith ("<"):
        return value

    return f"<{value}>"

def sparql_filter (name, value, escape=False, is_uri=False):
    """Returns a FILTER statement that can be added to a SPARQL query."""
    query   = ""
    if value is None:
        return query

    if is_uri:
        query  += f"FILTER (?{name} = {urify_value(value)})\n"
    else:
        # Wrapping the xsd:string in STR ensures compatibility with Virtuoso.
        literal = f"STR({escape_value (value, XSD.string)})" if escape else value
        symbol  = f"STR(?{name})" if escape else f"?{name}"
        query  += f"FILTER ({symbol}={literal})\n"

    return query

def sparql_contains_filter (names, values, lcase=True, andgate=False, decapsulate=False):
    """Returns a FILTER statement for CONTAINS() with multiple names and values."""
    query   = ""
    names  = [names]  if isinstance(names,  str) and names  != "" else names
    values = [values] if isinstance(values, str) and values != "" else values

    if not isinstance(names, list) or not isinstance(values, list):
        return query

    if not values or not names:
        return query

    operator = "&&" if andgate else "||"

    query = "FILTER (" if not decapsulate else ""
    for name in names:
        for value in values:
            symbol  = f"STR(?{name})"
            literal = f"STR({escape_value (value, XSD.string)})"
            if lcase:
                symbol  = f"LCASE(STR(?{name}))"
                literal = f"STR({escape_value (value.lower(), XSD.string)})"
            query  += f"CONTAINS({symbol},{literal}) {operator} "

    if not decapsulate:
        query = query[:-len(operator) - 2] + ")"
    else:
        query = query[:-len(operator) - 2]

    return query

def escape_value (value, datatype=None):
    """Returns VALUE wrapped in double quotes with type annotation DATATYPE."""
    if value is None:
        return None

    return Literal(value, datatype=datatype).n3()

def escape_string_value (value):
    """Returns VALUE wrapped in double quotes and annotated as xsd:string."""
    return escape_value (value, datatype=XSD.string)

def escape_date_value (value):
    """Returns VALUE wrapped in double quotes and annotated as xsd:date."""
    return escape_value (value, datatype=XSD.date)

def escape_datetime_value (value):
    """Returns VALUE wrapped in double quotes and annotated as xsd:dateTime."""
    return escape_value (value, datatype=XSD.dateTime)

def escape_boolean_value (value):
    """Returns VALUE wrapped in double quotes and annotated as xsd:boolean."""
    return escape_value (value, datatype=XSD.boolean)

def sparql_in_filter (name, values, escape=False, is_uri=False, negate=False):
    """Returns a FILTER statement for a list of values."""
    query   = ""
    # This is a deliberate loose check. We catch [], "", and None this way.
    if not values:
        return query

    compare = "NOT IN" if negate else "IN"
    if is_uri:
        query += f"FILTER (?{name} {compare} ({','.join(map(urify_value, values))}))\n"
    else:
        symbol = f"STR(?{name})" if escape else f"?{name}"
        escape_function = escape_string_value if escape else str
        query += f"FILTER (({symbol}) {compare} ({','.join(map(escape_function, values))}))\n"

    return query

def sparql_bound_filter (name):
    """Returns a FILTER statement to test whether a variable is BOUND."""
    return f"FILTER (BOUND(?{name}))\n"

def sparql_suffix (order, order_direction, limit=None, offset=None):
    """Returns a query suffix including order, limit and offset."""

    if order_direction is None:
        order_direction = "DESC"
    else:
        order_direction = order_direction.upper()

    query = ""
    if order and order_direction:
        if order[0] != '?':
            order = f"?{order}"
        query += f"ORDER BY {order_direction}({order})"

    if limit is not None:
        query += f"\nLIMIT {limit}"

    if offset is not None:
        query += f"\nOFFSET {offset}"

    return query

def insert_query (state_graph, rdflib_graph):
    """Procedure to generate a SPARQL query to insert triplets."""

    body  = rdflib_graph.serialize(format="ntriples")
    if isinstance(body, bytes):
        body = body.decode('utf-8')

    query = f"INSERT DATA {{ GRAPH <{state_graph}> {{ {body} }} }}"

    return query

def blank_node ():
    """Return a unique blank node."""
    identifier = str(uuid.uuid4())
    return BLANK[identifier]

def unique_node (prefix):
    """Return a unique node using PREFIX."""
    prefix_namespace = Namespace(f"{prefix}:")
    identifier       = str(uuid.uuid4())
    return prefix_namespace[identifier]

def uri_to_uuid (uri):
    """Returns the UUID of a URI created by 'unique_node' or 'blank_node'."""
    if uri is None:
        return None

    return uri[uri.find(":") + 1:]

def uuid_to_uri (uuid_value, datatype):
    """Returns a string of the full uri for UUID of type DATATYPE."""
    if uuid_value is None:
        return None

    return f"{datatype}:{uuid_value}"

def uris_from_records (records, prefix, uuid_index=None):
    """Returns URIRefs for a list of RECORDS of type PREFIX."""

    ## The UUID is sometimes a property of record, as in record["uuid"],
    ## while at other times the UUID is the record itself.

    if records and isinstance (records[0], URIRef):
        return records

    if records and isinstance(records[0], dict):
        if uuid_index is not None:
            return list(map (lambda record: URIRef(uuid_to_uri (
                record[uuid_index], prefix)), records))

    return list(map (lambda record: URIRef(
        uuid_to_uri (record, prefix)), records))



"""
This module implements constructing XML trees for rendering the DataCite
and other XML formats.
"""

from xml.etree import ElementTree
from djehuty.utils.convenience import value_or, value_or_none

class ElementMaker:
    '''
    Convenience class to simplify construction of xml trees. The number of
    steps is reduced by combining common ltree operations in a single call.
    An instance of this class can hold a set of namespace definitions,
    enabling tree construction using prefixes instead of full namespaces
    in element and attribute names. Default namespace is also obeyed.
    '''
    def __init__ (self, namespaces=None):
        self.namespaces = namespaces
        if self.namespaces is None:
            self.namespaces = {}

        for prefix, uri in self.namespaces.items():
            ElementTree.register_namespace(prefix, uri)

    def resolve (self, name, is_element=True):
        """Procedure to translate a prefixed NAME to its full namespace URI."""
        if ':' in name:
            prefix, suffix = name.split(':' ,1)
            if prefix != 'xml':
                return f'{{{self.namespaces[prefix]}}}{suffix}'

        if is_element and '' in self.namespaces:
            return f'{{{self.namespaces[""]}}}{name}'

        return name

    def child (self, parent, name, attrib=None, text=None):
        """Procedure to process a child element including attributes."""
        if parent is not None:
            element = ElementTree.SubElement(parent, self.resolve(name))
        else:
            element = ElementTree.Element(self.resolve(name))

        if attrib is not None:
            for attname, val in attrib.items():
                element.set(self.resolve(attname, False), val)

        if text:
            element.text = f"{text}"

        return element

    def child_option (self, parent, name, source, key, attrib=None):
        """Procedure to process a child element if KEY exists in SOURCE."""
        if key in source:
            return self.child(parent, name, attrib, f"{source[key]}")
        return None

    def root (self, name, attrib=None, schemas=None, text=None):
        """Procedure to make the ElementTree root."""
        attrib = attrib if attrib else {}
        if schemas:
            self.namespaces['xsi'] = 'http://www.w3.org/2001/XMLSchema-instance'
            ElementTree.register_namespace('xsi', self.namespaces['xsi'])
            schema_decl = ''.join(
                [f'{self.namespaces[pr]} {schema}' for pr, schema in schemas.items()])
            attrib['xsi:schemaLocation'] = schema_decl
        return self.child(None, name, attrib, text)

def serialize_tree_to_string (tree, indent=True):
    """Procedure to turn the ElementTree into a string."""
    if tree is None:
        return None
    if indent:
        ElementTree.indent(tree)
    return ElementTree.tostring(tree, encoding='utf8', short_empty_elements=True)

def scrub (obj):
    """Eliminate from construct of dicts and lists all values x for which bool(x)==False."""
    if isinstance(obj, dict):
        scrubbed = {key:scrub(val) for key,val in obj.items() if val}
        return {key:val for key,val in scrubbed.items() if val}

    if isinstance(obj, list):
        scrubbed = [scrub(val) for val in obj if val]
        return [val for val in scrubbed if val]

    return obj

def dublincore_tree (parameters):
    """Procedure to create a Dublin Core XML tree from PARAMETERS."""
    if parameters is None:
        return None
    parameters = scrub(parameters)
    namespaces = {'dc' : 'http://purl.org/dc/elements/1.1/',
                  'oai': 'http://www.openarchives.org/OAI/2.0/oai_dc/'}
    schemas =    {'oai': 'https://www.openarchives.org/OAI/2.0/oai_dc.xsd'}
    maker = ElementMaker(namespaces)
    root = maker.root('oai:dc', schemas=schemas)
    item = parameters['item']
    maker.child(root, 'dc:title', {}, item['title'])
    for creator in value_or (parameters, 'authors', []):
        maker.child_option(root, 'dc:creator', creator, 'full_name')
    for tag in value_or (parameters, 'tags', []):
        maker.child(root, 'dc:subject', {}, tag)
    maker.child(root, 'dc:description', {}, item['description'])
    maker.child(root, 'dc:publisher', {}, value_or(item, 'publisher', '4TU.ResearchData'))
    for contributor in value_or (parameters, 'contributor', []):
        maker.child(root, 'dc:contributor', {}, contributor['name'])
    for name in value_or (parameters, 'organizations', []):
        maker.child(root, 'dc:contributor', {}, name)
    if 'published_date' in parameters:
        maker.child(root, 'dc:date', {}, parameters['published_date'])
    maker.child(root, 'dc:type', {}, item['defined_type_name'])
    maker.child_option(root, 'dc:format', item, 'format')
    maker.child(root, 'dc:identifier', {}, parameters['doi'])
    maker.child(root, 'dc:language', {}, value_or(item, 'language', 'en'))
    if 'recource_doi' in item:
        maker.child(root, 'dc:relation', {}, f"https://doi.org/{item['recource_doi']}")
    maker.child_option(root, 'dc:coverage', item, 'geolocation')
    maker.child_option(root, 'dc:coverage', item, 'time_coverage')
    maker.child_option(root, 'dc:rights', item, 'license_name')
    return root

def dublincore (parameters):
    """Procedure to create a Dublin Core XML string from PARAMETERS."""
    return serialize_tree_to_string (dublincore_tree (parameters))

def nlm (parameters):
    """Procedure to create a NLM XML string from PARAMETERS."""
    if parameters is None:
        return None
    parameters = scrub(parameters)
    namespaces = {'xlink': 'http://www.w3.org/1999/xlink/'}
    maker = ElementMaker(namespaces)
    root    = maker.root('articles')
    article = maker.child(root, 'article')
    front   = maker.child(article, 'front')
    meta    = maker.child(front, 'article-meta')
    item = parameters['item']
    maker.child(maker.child(meta, 'title-group'),
                'article-title', {}, item['title'])
    contribs = maker.child(meta, 'contrib-group')
    if "authors" in parameters:
        for author in parameters['authors']:
            name = maker.child(maker.child(contribs, 'contrib', {'contrib-type': 'author'}), 'name')
            maker.child_option(name, 'surname', author, 'last_name')
            maker.child_option(name, 'given-name', author, 'first_name')
    maker.child(maker.child(meta, 'pub-date', {'pub-type': 'pub'}),
                'year', {}, value_or(parameters, 'published_year', None))
    maker.child(meta, 'self-uri', {'xlink:href': f"https:doi.org/{parameters['doi']}"})
    maker.child(front, 'abstract', {}, item['description'])
    return serialize_tree_to_string(root)

def refworks (parameters):
    """Procedure to create a Refworks XML string from PARAMETERS."""
    if parameters is None:
        return None
    parameters = scrub(parameters)
    maker = ElementMaker()
    root = maker.root('references')
    ref  = maker.child(root, 'reference')
    item    = parameters['item']
    if "authors" in parameters:
        authors = parameters['authors']
        for index, author in enumerate(authors):
            name = author['full_name']
            if 'last_name' in author and 'first_name' in author:
                name = f"{author['last_name']}, {author['first_name']}"
            maker.child(ref, f'a{index+1}', {}, name)
    maker.child(ref, 't1', {}, item['title'])
    maker.child(ref, 'sn')
    maker.child(ref, 'op')
    maker.child(ref, 'vo')
    maker.child(ref, 'ab', {}, item['description'])
    maker.child(ref, 'la', {}, value_or(item, 'language', 'en'))
    if 'tags' in parameters:
        for index, tag in enumerate(parameters['tags']):
            maker.child(ref, f'k{index+1}', {}, tag)
    maker.child(ref, 'pb')
    maker.child(ref, 'pp')
    maker.child(ref, 'yr', {}, value_or (parameters, 'published_year', 'unknown'))
    maker.child(ref, 'ed')
    doi = parameters['doi']
    maker.child(ref, 'ul', {}, f'https://doi.org/{doi}')
    maker.child(ref, 'do', {}, doi)
    return serialize_tree_to_string(root)

def datacite_tree (parameters, debug=False):
    """Procedure to create a DataCite XML tree from PARAMETERS."""
    if parameters is None:
        return None
    parameters = scrub(parameters)
    namespaces = {'': 'http://datacite.org/schema/kernel-4'}
    schemas    = {'': 'http://schema.datacite.org/meta/kernel-4.4/metadata.xsd'}
    maker = ElementMaker(namespaces)
    root = maker.root('resource', schemas=schemas)
    item = parameters['item']

    #01 identifier
    maker.child(root, 'identifier', {'identifierType':'DOI'}, parameters['doi'])

    #02 creators
    orcid_att = {'nameIdentifierScheme': 'https://orcid.org/'}
    personal_att = {'nameType': 'Personal'}
    if "authors" in parameters:
        creators = parameters['authors']
        creators_element = maker.child(root,'creators')
        for creator in creators:
            creator_att = personal_att if 'orcid_id' in creator else {}
            creator_element = maker.child(creators_element, 'creator')
            maker.child_option(creator_element, 'creatorName', creator, 'full_name', creator_att)
            maker.child_option(creator_element, 'givenName', creator, 'first_name')
            maker.child_option(creator_element, 'familyName', creator, 'last_name')
            maker.child_option(creator_element, 'nameIdentifier', creator, 'orcid_id', orcid_att)

    #03 titles
    maker.child(maker.child(root, 'titles'), 'title', {}, item['title'])

    #04 publisher
    maker.child(root, 'publisher', {}, value_or(item, 'publisher', '4TU.ResearchData'))

    #05 publicationYear
    maker.child(root, 'publicationYear', {}, value_or(parameters, 'published_year', None))

    #06 resourceType
    rtype = value_or(item, 'defined_type_name', 'Collection').capitalize()
    if rtype not in ('Dataset','Software','Collection'):
        rtype = 'Text'
    maker.child(root, 'resourceType', {'resourceTypeGeneral': rtype}, rtype)

    #07 subjects
    subjects_element = maker.child(root, 'subjects')
    if 'categories' in parameters:
        for cat in parameters['categories']:
            maker.child(
                subjects_element,
                'subject',
                { 'subjectScheme'     :
                  'Australian and New Zealand Standard Research Classification (ANZSRC), 2008',
                  'classificationCode': cat['classification_code'] },
                cat['title']
            )
    if 'tags' in parameters:
        for tag in parameters['tags']:
            maker.child(subjects_element, 'subject', {}, tag)
    if 'time_coverage' in item:
        maker.child(subjects_element, 'subject', {}, f"Time: {item['time_coverage']}")

    #08 contributors
    has_organizations = 'organizations' in parameters
    has_contributors = 'contributors' in parameters
    if has_organizations or has_contributors:
        contributors_element = maker.child(root, 'contributors')
        type_att = {'contributorType': 'Other'}
        orcid_att = {'nameIdentifierScheme': 'https://orcid.org/'}
        if has_contributors:
            for contributor in parameters['contributors']:
                contributor_element = maker.child(contributors_element, 'contributor', type_att)
                name = contributor['name']
                orcid = value_or_none(contributor, 'orcid')
                maker.child(contributor_element, 'contributorName', {'nameType': 'Personal'}, name)
                if orcid:
                    maker.child(contributor_element, 'nameIdentifier', orcid_att, orcid)
        if has_organizations:
            for name in parameters['organizations']:
                contributor_element = maker.child(contributors_element, 'contributor', type_att)
                maker.child(contributor_element, 'contributorName',
                            {'nameType': 'Organizational'}, name)

    #09 dates
    if 'published_date' in parameters:
        maker.child(maker.child(root, 'dates'), 'date',
                    {'dateType': 'Issued'}, parameters['published_date'])

    #10 language
    if 'language' in item:
        maker.child(root, 'language', {}, item['language'])

    #11 relatedIdentifiers
    has_resource_doi = 'resource_doi' in item
    has_ref = 'references' in parameters
    has_other_versions = "version" in item and "container_doi" in parameters
    if has_resource_doi or has_ref or has_other_versions:
        relations_element = maker.child(root, 'relatedIdentifiers')
        if has_resource_doi:
            maker.child(relations_element,
                        'relatedIdentifier',
                        {'relatedIdentifierType': 'DOI', 'relationType': 'IsSupplementTo'},
                        item['resource_doi'])
        if has_ref:
            for ref in parameters['references']:
                maker.child(relations_element,
                            'relatedIdentifier',
                            {'relatedIdentifierType': 'URL', 'relationType': 'References'},
                            ref['url'])
        if has_other_versions:
            version = int(item['version'])
            for index in range(1, version + 1):
                versioned_doi = f"{parameters['container_doi']}.v{index}"
                if parameters['doi'] != versioned_doi:
                    maker.child(relations_element, 'relatedIdentifier',
                                { 'relatedIdentifierType': 'DOI', 'relationType': 'IsVersionOf' },
                                versioned_doi)

    #12 formats
    if 'format' in item:
        maker.child(maker.child(root, 'formats'), 'format', {}, item['format'])

    #13 version
    maker.child(root, 'version', {}, f'{item["version"]}')

    #14 rightsList
    if 'license_id' in item:
        maker.child(maker.child(root, 'rightsList'),
                    'rights',
                    {'rightsURI': item['license_url']},
                    item['license_name'])

    #15 descriptions
    maker.child(maker.child(root, 'descriptions'),
                'description',
                {'descriptionType': 'Abstract'},
                item['description'])

    #16 geoLocations
    has_geo = 'geolocation' in item
    coordinates = value_or(parameters, 'coordinates', {})
    has_point = 'lat_valid' in coordinates and 'lon_valid' in coordinates
    if has_geo or has_point:
        geo_element = maker.child(maker.child(root, 'geoLocations'), 'geoLocation')
        if has_geo:
            maker.child(geo_element, 'geoLocationPlace', {}, item['geolocation'])
        if has_point:
            point_element = maker.child(geo_element, 'geoLocationPoint')
            maker.child(point_element, 'pointLongitude', {}, coordinates['lon_valid'])
            maker.child(point_element, 'pointLatitude', {}, coordinates['lat_valid'])

    #17 fundingReferences
    if 'fundings' in parameters:
        fundings_element = maker.child(root, 'fundingReferences')
        for funding in parameters['fundings']:
            funding_element = maker.child(fundings_element, 'fundingReference')
            maker.child(funding_element, 'funderName', {},
                        value_or(funding, 'funder_name', 'unknown'))
            maker.child_option(funding_element, 'awardNumber', funding, 'grant_code')
            maker.child_option(funding_element, 'awardTitle', funding, 'title')

    #debug
    if debug:
        param_strings = [f'{key:<15}: {val}' for key, val in parameters.items()]
        root.insert(0, ElementTree.Comment('DEBUG\n' + '\n'.join(param_strings)))

    return root

def datacite (parameters, indent=True):
    """Procedure to create a DataCite XML string from PARAMETERS.
       For registration at Datacite, set indent to False"""
    return serialize_tree_to_string (datacite_tree(parameters), indent=indent)



"""This module implements interaction with an S3 endpoint."""

import logging
import uuid
import os
from datetime import datetime
from djehuty.web.config import config
from djehuty.utils.convenience import value_or

try:
    import boto3
    from botocore.exceptions import ClientError, PartialCredentialsError
    from botocore.exceptions import ResponseStreamingError, ReadTimeoutError
    from botocore.config import Config
    from urllib3.exceptions import IncompleteRead
except (ImportError, ModuleNotFoundError):
    pass

class S3DownloadStreamer:
    """Generator to stream the contents of a file stored in S3."""

    def __init__ (self, endpoint, bucket, access_key, secret_key, filename, name, chunk_size=32768, offset=0):
        self.client = None
        self.endpoint = endpoint
        self.bucket = bucket
        self.access_key = access_key
        self.secret_key = secret_key
        self.filename = filename
        self.chunk_size = chunk_size
        self.offset = offset
        self.log = logging.getLogger (__name__)
        self.chunk_size = chunk_size
        self.original_filename = name
        self.content_length = 0
        self.content_type = "binary/octet-stream"
        self.last_modified = None
        self.file_object = None
        self.file_contents = None
        self.boto_config = Config(retries = { "total_max_attempts": 30,
                                              "mode": "standard" },
                                  max_pool_connections = 1,
                                  read_timeout = 120)

    def connect (self):
        """Initialize procedure that can be recalled."""
        self.client = boto3.client ("s3", endpoint_url    = self.endpoint,
                                    aws_access_key_id     = self.access_key,
                                    aws_secret_access_key = self.secret_key,
                                    config                = self.boto_config)
        try:
            self.file_object   = self.client.get_object (Bucket = self.bucket,
                                                         Key    = self.filename,
                                                         Range  = f"bytes={self.offset}-")
            self.file_contents = self.file_object["Body"]
        except (ClientError, KeyError) as error:
            self.log.error ("An S3 download stream error occurred: %s", error)

        try:
            http_headers = self.file_object["ResponseMetadata"]["HTTPHeaders"]
            self.content_type   = value_or (http_headers, "content-type", self.content_type)
            self.content_length = int(value_or (http_headers, "content-length", 0))
            modified = datetime.strptime (value_or (http_headers, "last-modified",
                                                    "Tue, 01 Jan 1980 12:00:00 GMT"),
                                          "%a, %d %b %Y %H:%M:%S %Z")
            self.last_modified  = (modified.year, modified.month, modified.day,
                                   modified.hour, modified.minute, modified.second)
        except (KeyError, TypeError):
            self.log.warning ("Could not read metadata for s3://%s/%s",
                              self.bucket, self.filename)

    def body (self):
        """Returns the request body to directly read from."""
        if self.file_contents is None:
            self.connect ()

        return self.file_contents

    def iterator (self):
        """Returns an iterator to read the request body."""
        if self.file_contents is None:
            self.connect ()

        return self.file_contents.iter_chunks (chunk_size=self.chunk_size)

    def close (self):
        """Closes the S3 client and resets the internal state."""
        self.file_contents.close()
        self.client.close()
        self.file_object = None
        self.file_contents = None
        self.content_length = 0
        self.content_type = "binary/octet-stream"
        self.last_modified = None

    def reset (self, offset=0):
        """Resets the S3 connection and ttempt to contunie reading at OFFSET."""
        self.close ()
        self.offset = offset
        self.connect ()

def s3_file_exists (endpoint, bucket, access_key, secret_key, filename):
    """Returns True when FILENAME exists in BUCKET, False otherwise."""
    try:
        client = boto3.client("s3", endpoint_url=endpoint,
                              aws_access_key_id=access_key,
                              aws_secret_access_key=secret_key)
        client.head_object (Bucket=bucket, Key=filename)
        return True
    except PartialCredentialsError:
        logger = logging.getLogger(__name__)
        logger.warning ("Potential misconfiguration of S3 bucket '%s'.", bucket)
        return False
    except ClientError:
        return False

def s3_temporary_file (reader):
    """Downloads the S3 file from READER and returns the local filesystem path."""
    cached_filename = os.path.join (config.s3_cache_storage, str(uuid.uuid4()))
    with open (cached_filename, "wb") as output_stream:
        retries = 3
        while retries > 0:
            try:
                for chunk in reader.iterator():
                    output_stream.write (chunk)
                retries = 0
            except (ResponseStreamingError, ReadTimeoutError, IncompleteRead):
                logger = logging.getLogger (__name__)
                current_offset = reader.body().tell()
                reader.reset (offset = current_offset)
                retries -= 1
                if retries > 0:
                    logger.warning ("Retrying to fetch after %s bytes of %s.",
                                    current_offset, reader.original_filename)
                    continue
                logger.error ("Failed to fetch S3 object %s (%s) for ZIP.",
                              reader.original_filename,
                              reader.content_length)
    reader.close()
    return cached_filename
